@@ -13,6 +13,16 @@
 
 (in-package "SB-IMPL")
 
+;;; Default evaluator mode (interpreter / compiler)
+
+(declaim (type (member :compile #+(or sb-eval sb-fasteval) :interpret)
+               *evaluator-mode*))
+(defparameter *evaluator-mode* :compile
+  "Toggle between different evaluator implementations. If set to :COMPILE,
+an implementation of EVAL that calls the compiler will be used. If set
+to :INTERPRET, an interpreter will be used.")
+(declaim (always-bound *evaluator-mode*))
+
 ;;; Like DEFUN, but hides itself in the backtrace. This is meant for
 ;;; trivial functions which just do some argument parsing and call
 ;;; ERROR for real. Hence, having them and their locals in the
@@ -224,22 +234,27 @@
   (let ((macros ())
         (binds ())
         (dx ())
-        (ignores ()))
+        (ignores ())
+        (declares))
     (dolist (spec collections)
       (destructuring-bind (name &optional initial-value (collector nil collectorp)) spec
         (multiple-value-bind (name append-tail)
             (if (consp name)
                 (values (first name) (second name))
                 (values name nil))
-          (let ((n-value (copy-symbol name)))
-            (push `(,n-value ,(if (or initial-value collectorp)
-                                  initial-value
-                                  `(#-sb-xc-host unaligned-dx-cons
-                                    #+sb-xc-host list
-                                    nil)))
+          (let ((n-value (copy-symbol name))
+                (n-tail (gensymify* name "-TAIL")))
+            (push `(,n-value ,(cond ((or initial-value collectorp)
+                                     initial-value)
+                                    (t
+                                     #-sb-xc-host (push `(sb-c::no-debug ,n-value ,n-tail) declares)
+                                     `(#-sb-xc-host unaligned-dx-cons
+                                       #+sb-xc-host list
+                                       nil))))
                   binds)
-            (let* ((n-tail (gensymify* name "-TAIL"))
-                   (macro-body
+            ;; Hide unaligned-dx-cons
+
+            (let* ((macro-body
                      (cond
                        (collectorp
                         ``(progn
@@ -274,7 +289,8 @@
                 (push `(,append-tail (x) `(setf (cdr ,',n-tail) ,x)) macros)))))))
     `(macrolet ,macros
        (let* ,(nreverse binds)
-         ,@(if dx `((declare (dynamic-extent ,@dx))))
+         ,@(if dx `((declare (dynamic-extent ,@dx)
+                             ,@declares)))
          ;; Even if the user reads each collection result,
          ;; reader conditionals might statically eliminate all writes.
          ;; Since we don't know, all the -n-tail variable are ignorable.
@@ -516,8 +532,7 @@ NOTE: This interface is experimental and subject to change."
   (dolist (name *cache-vector-symbols*)
     ;; We really don't need to call ABOUT-TO-MODIFY-SYMBOL-VALUE 18 times
     ;; just to clear the caches.
-    #-sb-xc-host (%primitive %set-symbol-global-value name nil)
-    #+sb-xc-host (set name nil)))
+    (%set-symbol-global-value name nil)))
 
 (defmacro sb-int-package () (find-package "SB-INT"))
 
@@ -528,10 +543,7 @@ NOTE: This interface is experimental and subject to change."
   (let (cache)
     ;; It took me a while to figure out why infinite recursion could occur
     ;; in VALUES-SPECIFIER-TYPE. It's because SET calls VALUES-SPECIFIER-TYPE.
-    (macrolet ((set! (symbol value)
-                 #-sb-xc-host `(%primitive %set-symbol-global-value ,symbol ,value)
-                 #+sb-xc-host `(set ,symbol ,value))
-               (reset-stats ()
+    (macrolet ((reset-stats ()
                  ;; If statistics gathering is not not compiled-in,
                  ;; no sense in setting a symbol that is never used.
                  ;; While this uses SYMBOLICATE at runtime,
@@ -540,7 +552,7 @@ NOTE: This interface is experimental and subject to change."
                      `(let ((statistics
                              (package-symbolicate (sb-int-package) symbol "-STATS")))
                         (unless (boundp statistics)
-                          (set! statistics
+                          (%set-symbol-global-value statistics
                                 (make-array 3 :element-type 'fixnum
                                               :initial-contents '(1 0 0))))))))
       ;; It would be bad if another thread sees MAKE-ARRAY's result in the
@@ -550,7 +562,7 @@ NOTE: This interface is experimental and subject to change."
       (sb-thread:barrier (:write)
         (reset-stats)
         (setq cache (make-array size :initial-element 0)))
-      (set! symbol cache)
+      (%set-symbol-global-value symbol cache)
       cache)))
 
 ;; At present we make a new vector every time a line is re-written,
@@ -1312,7 +1324,8 @@ NOTE: This interface is experimental and subject to change."
                                      (memq id '(special ignorable ignore
                                                 dynamic-extent
                                                 sb-c::constant-value
-                                                sb-c::no-constraints))
+                                                sb-c::no-constraints
+                                                sb-c::no-debug))
                                      (info :type :kind id))
                                  (cdr decl))))))
            (partition (spec)
@@ -1444,6 +1457,7 @@ NOTE: This interface is experimental and subject to change."
 
 (in-package "SB-KERNEL")
 
+(declaim (inline fp-zero-p))
 (defun fp-zero-p (x)
   (typecase x
     (single-float (zerop x))
