@@ -18,10 +18,14 @@
 
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (defun handle-reals (function var)
-    `((((foreach fixnum single-float bignum ratio))
-       (coerce (,function (coerce ,var 'double-float)) 'single-float))
-      ((double-float)
-       (,function ,var))))
+    (let ((f (find-symbol (format nil "~aF" function))))
+      (aver f)
+      `((((foreach fixnum bignum ratio))
+         (,f (coerce ,var 'single-float)))
+        ((single-float)
+         (,f ,var))
+        ((double-float)
+         (,function ,var)))))
 
   (defun handle-complex (form)
     `((((foreach (complex double-float) (complex single-float) (complex rational)))
@@ -29,57 +33,61 @@
 
 ;;; Make these INLINE, since the call to C is at least as compact as a
 ;;; Lisp call, and saves number consing to boot.
-(defmacro def-math-rtn (name num-args &optional wrapper)
-  (let ((function (symbolicate "%" (string-upcase name)))
-        (args (loop for i below num-args
-                    collect (intern (format nil "ARG~D" i)))))
+(defmacro def-math-rtn (name num-args &optional wrapper transform)
+  (flet ((make-fun (dname &optional f)
+           (let* ((name (if f
+                            (format nil "~af" dname)
+                            dname))
+                  (function (symbolicate "%" (string-upcase name)))
+                  (args (loop for i below num-args
+                              collect (intern (format nil "ARG~D" i))))
+                  (type (if f
+                            'single-float
+                            'double-float))
+                  (body
+                    (cond
+                      #-libmf
+                      (f
+                       `(coerce (,(symbolicate "%" (string-upcase dname))
+                                 ,@(loop for arg in args
+                                         collect `(coerce ,arg 'double-float)))
+                                'single-float))
+                      (t
+                       `(locally
+                            (declare (sb-c:flushable sb-c:%alien-funcall))
+                          (truly-the ;; avoid checking the result
+                           ,(type-specifier (fun-type-returns (info :function :type function)))
+                           (alien-funcall
+                            (extern-alien ,(format nil "~:[~;sb_~]~a" (and wrapper #+ucrt nil) name)
+                                          (function ,type
+                                                    ,@(loop repeat num-args
+                                                            collect type)))
+                            ,@args)))))))
+             (if transform
+                 `(sb-c:deftransform ,function (,args)
+                    ',body)
+                 `(progn
+                    (declaim (inline ,function))
+                    (defun ,function ,args
+                      ,body))))))
     `(progn
-       (declaim (inline ,function))
-       (defun ,function ,args
-         (declare (sb-c:flushable sb-c:%alien-funcall))
-         (truly-the ;; avoid checking the result
-          ,(type-specifier (fun-type-returns (info :function :type function)))
-          (alien-funcall
-           (extern-alien ,(format nil "~:[~;sb_~]~a" wrapper name)
-                         (function double-float
-                                   ,@(loop repeat num-args
-                                           collect 'double-float)))
-           ,@args))))))
-
+       ,(make-fun name)
+       ,(make-fun name t))))
 
-#+x86 ;; for constant folding
+#+(or #+x86 x86-64 arm-vfp arm64 riscv) ;; for constant folding
 (macrolet ((def (name ll)
              `(defun ,name ,ll (,name ,@ll))))
-  (def %atan2 (x y))
-  (def %atan (x))
-  (def %tan (x))
-  (def %tan-quick (x))
-  (def %cos (x))
-  (def %cos-quick (x))
-  (def %sin (x))
-  (def %sin-quick (x))
   (def %sqrt (x))
-  (def %log (x))
-  (def %log1p (x))
-  (def %log2 (x))
-  (def %exp (x)))
-
-#+(or x86-64 arm-vfp arm64 riscv) ;; for constant folding
-(macrolet ((def (name ll)
-             `(defun ,name ,ll (,name ,@ll))))
-  (def %sqrt (x)))
+  (def %sqrtf (x)))
 
 ;;;; stubs for the Unix math library
-;;;;
-;;;; Many of these are unnecessary on the X86 because they're built
-;;;; into the FPU.
 
 ;;; trigonometric
-#-x86 (def-math-rtn "sin" 1)
-#-x86 (def-math-rtn "cos" 1)
-#-x86 (def-math-rtn "tan" 1)
-#-x86 (def-math-rtn "atan" 1)
-#-x86 (def-math-rtn "atan2" 2)
+(def-math-rtn "sin" 1)
+(def-math-rtn "cos" 1)
+(def-math-rtn "tan" 1)
+(def-math-rtn "atan" 1)
+(def-math-rtn "atan2" 2)
 
 ;;; See src/runtime/wrap.c for the definitions of the "sb_"-prefixed things.
 (def-math-rtn "acos" 1 #+win32 t)
@@ -93,13 +101,14 @@
 
 ;;; exponential and logarithmic
 (def-math-rtn "hypot" 2 #+win32 t)
-#-x86 (def-math-rtn "exp" 1)
-#-x86 (def-math-rtn "log" 1)
-#-x86 (def-math-rtn "log10" 1)
+(def-math-rtn "exp" 1)
+(def-math-rtn "log" 1)
+(def-math-rtn "log10" 1)
 (def-math-rtn "pow" 2)
-#-(or x86 x86-64 arm-vfp arm64 riscv) (def-math-rtn "sqrt" 1)
-#-x86 (def-math-rtn "log1p" 1)
-#-x86 (def-math-rtn "log2" 1)
+#-(or x86 x86-64 arm-vfp arm64 riscv)
+(def-math-rtn "sqrt" 1 nil #+ppc64 t)
+(def-math-rtn "log1p" 1)
+(def-math-rtn "log2" 1)
 
 
 ;;;; power functions
@@ -194,6 +203,7 @@
          (locally (declare (notinline expt))
            (expt 10 power)))))
 
+(declaim (maybe-inline expt))
 ;;; If an integer power of a rational, use INTEXP above. Otherwise, do
 ;;; floating point stuff. If both args are real, we try %POW right
 ;;; off, assuming it will return 0 if the result may be complex. If
@@ -202,184 +212,168 @@
 ;;; from the general complex case.
 (defun expt (base power)
   "Return BASE raised to the POWER."
-  (declare (explicit-check))
-  (if (zerop power)
-    (if (and (zerop base) (floatp power))
-        (error 'arguments-out-of-domain-error
-               :operands (list base power)
-               :operation 'expt
-               :references '((:ansi-cl :function expt)))
-        (let ((result (1+ (* base power))))
-          (if (and (floatp result) (float-nan-p result))
-              (float 1 result)
-              result)))
-    (labels (;; determine if the double float is an integer.
-             ;;  0 - not an integer
-             ;;  1 - an odd int
-             ;;  2 - an even int
-             (isint (ihi lo)
-               (declare (type (unsigned-byte 31) ihi)
-                        (type (unsigned-byte 32) lo)
-                        (optimize (speed 3) (safety 0)))
-               (let ((isint 0))
-                 (declare (type fixnum isint))
-                 (cond ((>= ihi #x43400000)     ; exponent >= 53
-                        (setq isint 2))
-                       ((>= ihi #x3ff00000)
-                        (let ((k (- (ash ihi -20) #x3ff)))      ; exponent
-                          (declare (type (mod 53) k))
-                          (cond ((> k 20)
-                                 (let* ((shift (- 52 k))
-                                        (j (logand (ash lo (- shift))))
-                                        (j2 (ash j shift)))
-                                   (declare (type (mod 32) shift)
-                                            (type (unsigned-byte 32) j j2))
-                                   (when (= j2 lo)
-                                     (setq isint (- 2 (logand j 1))))))
-                                ((= lo 0)
-                                 (let* ((shift (- 20 k))
-                                        (j (ash ihi (- shift)))
-                                        (j2 (ash j shift)))
-                                   (declare (type (mod 32) shift)
-                                            (type (unsigned-byte 31) j j2))
-                                   (when (= j2 ihi)
-                                     (setq isint (- 2 (logand j 1))))))))))
-                 isint))
-             (real-expt (x y rtype)
-               (let ((x (coerce x 'double-float))
-                     (y (coerce y 'double-float)))
-                 (declare (double-float x y))
-                 (let* ((x-hi (double-float-high-bits x))
-                        (x-lo (double-float-low-bits x))
-                        (x-ihi (logand x-hi #x7fffffff))
-                        (y-hi (double-float-high-bits y))
-                        (y-lo (double-float-low-bits y))
-                        (y-ihi (logand y-hi #x7fffffff)))
-                   (declare (type (signed-byte 32) x-hi y-hi)
-                            (type (unsigned-byte 31) x-ihi y-ihi)
-                            (type (unsigned-byte 32) x-lo y-lo))
-                   ;; y==zero: x**0 = 1
-                   (when (zerop (logior y-ihi y-lo))
-                     (return-from real-expt (coerce 1d0 rtype)))
-                   ;; +-NaN return x+y
-                   ;; FIXME: Hardcoded qNaN/sNaN values are not portable.
-                   (when (or (> x-ihi #x7ff00000)
-                             (and (= x-ihi #x7ff00000) (/= x-lo 0))
-                             (> y-ihi #x7ff00000)
-                             (and (= y-ihi #x7ff00000) (/= y-lo 0)))
-                     (return-from real-expt (coerce (+ x y) rtype)))
-                   (let ((yisint (if (< x-hi 0) (isint y-ihi y-lo) 0)))
-                     (declare (type fixnum yisint))
-                     ;; special value of y
-                     (when (and (zerop y-lo) (= y-ihi #x7ff00000))
-                       ;; y is +-inf
-                       (return-from real-expt
-                         (cond ((and (= x-ihi #x3ff00000) (zerop x-lo))
-                                ;; +-1**inf is NaN
-                                (coerce (- y y) rtype))
-                               ((>= x-ihi #x3ff00000)
-                                ;; (|x|>1)**+-inf = inf,0
-                                (if (>= y-hi 0)
-                                    (coerce y rtype)
-                                    (coerce 0 rtype)))
-                               (t
-                                ;; (|x|<1)**-,+inf = inf,0
-                                (if (< y-hi 0)
-                                    (coerce (- y) rtype)
-                                    (coerce 0 rtype))))))
+  (declare (explicit-check)
+           (maybe-inline expt))
+  (labels (;; determine if the double float is an integer
+           #+64-bit
+           (isint (n)
+             (or (zerop n)
+                 (let* ((bits (sb-kernel:double-float-bits n))
+                        (exponent (ldb sb-vm:double-float-exponent-byte bits)))
+                   (cond
+                     ((> exponent (+ sb-vm:double-float-bias sb-vm:double-float-digits)))
+                     ((> exponent sb-vm:double-float-bias)
+                      (not (logtest bits
+                                    (1- (ash 1 (- (+ sb-vm:double-float-bias sb-vm:double-float-digits)
+                                                  exponent))))))))))
+           #-64-bit
+           (isint (x)
+             (or (zerop x)
+                 (let* ((hi (double-float-high-bits x))
+                        (lo (double-float-low-bits x))
+                        (ihi (logand hi #x7fffffff)))
+                   (declare (type (unsigned-byte 31) ihi)
+                            (type (unsigned-byte 32) lo)
+                            (optimize (speed 3) (safety 0)))
+                   (let ((isint 0))
+                     (declare (type fixnum isint))
+                     (cond ((>= ihi #x43400000) ; exponent >= 53
+                            (setq isint 2))
+                           ((>= ihi #x3ff00000)
+                            (let ((k (- (ash ihi -20) #x3ff))) ; exponent
+                              (declare (type (mod 53) k))
+                              (cond ((> k 20)
+                                     (let* ((shift (- 52 k))
+                                            (j (logand (ash lo (- shift))))
+                                            (j2 (ash j shift)))
+                                       (declare (type (mod 32) shift)
+                                                (type (unsigned-byte 32) j j2))
+                                       (when (= j2 lo)
+                                         (setq isint (- 2 (logand j 1))))))
+                                    ((= lo 0)
+                                     (let* ((shift (- 20 k))
+                                            (j (ash ihi (- shift)))
+                                            (j2 (ash j shift)))
+                                       (declare (type (mod 32) shift)
+                                                (type (unsigned-byte 31) j j2))
+                                       (when (= j2 ihi)
+                                         (setq isint (- 2 (logand j 1))))))))))
+                     (/= isint 0)))))
+           (real-expt (x y rtype)
+             (let ((x (coerce x 'double-float))
+                   (y (coerce y 'double-float)))
+               (cond ((or (>= x 0)
+                          (= x double-float-negative-infinity)
+                          (isint y))
+                      (coerce (%pow x y) rtype))
+                     (t
+                      (let ((y*pi (* y pi))
+                            (pow (%pow (- x) y)))
+                        (complex
+                         (coerce (* pow (%cos y*pi)) rtype)
+                         (coerce (* pow (%sin y*pi)) rtype)))))))
+           (complex-expt (base power)
+             (if (zerop power)
+                 (1+ (* base power))
+                 (if (and (zerop base) (plusp (realpart power)))
+                     (* base power)
+                     (exp (* power (log base)))))))
+    (declare (inline real-expt complex-expt isint))
+    (number-dispatch ((base number) (power number))
+      (((foreach fixnum (or bignum ratio) (complex rational)) integer)
+       (intexp base power))
+      (((foreach single-float double-float) (foreach fixnum bignum))
+       (expt base power))
+      (((foreach double-float) ratio)
+       (real-expt base power '(dispatch-type base)))
+      (((foreach fixnum (or bignum ratio) single-float)
+        (foreach ratio single-float))
+       (real-expt base power 'single-float))
+      (((foreach fixnum (or bignum ratio) single-float double-float)
+        double-float)
+       (real-expt base power 'double-float))
+      ((double-float single-float)
+       (real-expt base power 'double-float))
+      ;; Handle (expt <complex> <rational>), except the case dealt with
+      ;; in the first clause above, (expt <(complex rational)> <integer>).
+      (((complex rational)
+        ratio)
+       (* (expt (abs base) power)
+          (cis (* power (phase base)))))
+      (((foreach (complex single-float) (complex double-float))
+        (foreach fixnum (or bignum ratio)))
+       (if (zerop power)
+           (1+ (* base 0))
+           (* (expt (abs base) power)
+              (cis (* power (phase base))))))
+      ;; The next three clauses handle (expt <real> <complex>).
+      (((foreach fixnum (or bignum ratio) single-float)
+        (foreach (complex single-float) (complex rational)))
+       (complex-expt base power))
+      (((foreach fixnum (or bignum ratio) single-float)
+        (complex double-float))
+       (complex-expt (coerce base 'double-float) power))
+      ((double-float complex)
+       (complex-expt base power))
+      ;; The next three clauses handle (expt <complex> <float>) and
+      ;; (expt <complex> <complex>).
+      (((foreach (complex single-float) (complex rational))
+        (foreach (complex single-float) (complex rational) single-float))
+       (complex-expt base power))
+      (((foreach (complex single-float) (complex rational))
+        (foreach (complex double-float) double-float))
+       (complex-expt (coerce base '(complex double-float)) power))
+      (((complex double-float)
+        (foreach complex double-float single-float))
+       (complex-expt base power)))))
 
-                     (let ((abs-x (abs x)))
-                       (declare (double-float abs-x))
-                       ;; special value of x
-                       (when (and (zerop x-lo)
-                                  (or (= x-ihi #x7ff00000) (zerop x-ihi)
-                                      (= x-ihi #x3ff00000)))
-                         ;; x is +-0,+-inf,+-1
-                         (let ((z (if (< y-hi 0)
-                                      (/ 1 abs-x)       ; z = (1/|x|)
-                                      abs-x)))
-                           (declare (double-float z))
-                           (when (< x-hi 0)
-                             (cond ((and (= x-ihi #x3ff00000) (zerop yisint))
-                                    ;; (-1)**non-int
-                                    (let ((y*pi (* y pi)))
-                                      (declare (double-float y*pi))
-                                      (return-from real-expt
-                                        (complex
-                                         (coerce (%cos y*pi) rtype)
-                                         (coerce (%sin y*pi) rtype)))))
-                                   ((= yisint 1)
-                                    ;; (x<0)**odd = -(|x|**odd)
-                                    (setq z (- z)))))
-                           (return-from real-expt (coerce z rtype))))
+(defun expt-double-float (base power)
+  (declare (inline expt)
+           (double-float base power))
+  (expt base power))
 
-                       (if (>= x-hi 0)
-                           ;; x>0
-                           (coerce (%pow x y) rtype)
-                           ;; x<0
-                           (let ((pow (%pow abs-x y)))
-                             (declare (double-float pow))
-                             (case yisint
-                               (1 ; odd
-                                (coerce (* -1d0 pow) rtype))
-                               (2 ; even
-                                (coerce pow rtype))
-                               (t ; non-integer
-                                (let ((y*pi (* y pi)))
-                                  (declare (double-float y*pi))
-                                  (complex
-                                   (coerce (* pow (%cos y*pi))
-                                           rtype)
-                                   (coerce (* pow (%sin y*pi))
-                                           rtype))))))))))))
-             (complex-expt (base power)
-               (if (and (zerop base) (plusp (realpart power)))
-                   (* base power)
-                   (exp (* power (log base))))))
-      (declare (inline real-expt complex-expt))
-      (number-dispatch ((base number) (power number))
-        (((foreach fixnum (or bignum ratio) (complex rational)) integer)
-         (intexp base power))
-        (((foreach single-float double-float) rational)
-         (real-expt base power '(dispatch-type base)))
-        (((foreach fixnum (or bignum ratio) single-float)
-          (foreach ratio single-float))
-         (real-expt base power 'single-float))
-        (((foreach fixnum (or bignum ratio) single-float double-float)
-          double-float)
-         (real-expt base power 'double-float))
-        ((double-float single-float)
-         (real-expt base power 'double-float))
-        ;; Handle (expt <complex> <rational>), except the case dealt with
-        ;; in the first clause above, (expt <(complex rational)> <integer>).
-        (((foreach (complex rational))
-          ratio)
-         (* (expt (abs base) power)
-            (cis (* power (phase base)))))
-        (((foreach (complex single-float) (complex double-float))
-          (foreach fixnum (or bignum ratio)))
-         (* (expt (abs base) power)
-            (cis (* power (phase base)))))
-        ;; The next three clauses handle (expt <real> <complex>).
-        (((foreach fixnum (or bignum ratio) single-float)
-          (foreach (complex single-float) (complex rational)))
-         (complex-expt base power))
-        (((foreach fixnum (or bignum ratio) single-float)
-          (complex double-float))
-         (complex-expt (coerce base 'double-float) power))
-        ((double-float complex)
-         (complex-expt base power))
-        ;; The next three clauses handle (expt <complex> <float>) and
-        ;; (expt <complex> <complex>).
-        (((foreach (complex single-float) (complex rational))
-          (foreach (complex single-float) (complex rational) single-float))
-         (complex-expt base power))
-        (((foreach (complex single-float) (complex rational))
-          (foreach (complex double-float) double-float))
-         (complex-expt (coerce base '(complex double-float)) power))
-        (((complex double-float)
-          (foreach complex double-float single-float))
-         (complex-expt base power))))))
+(defun expt-single-float (base power)
+  (declare (inline expt)
+           (single-float base power))
+  (expt base power))
+
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (dolist (s '(expt))
+    (clear-info :function :inlining-data s)
+    (clear-info :function :inlinep s)
+    (clear-info :source-location :declaration s)))
+
+(defun sqrt-double-float (number)
+  (if (< number 0)
+      (complex 0d0 (sqrt (- number)))
+      (sqrt number)))
+
+(defun sqrt-single-float (number)
+  (if (< number 0)
+      (complex 0f0 (sqrt (- number)))
+      (sqrt number)))
+
+(defun log-double-float (number)
+  (if (< number 0)
+      (complex (log (- number)) pi)
+      (log number)))
+
+(defun log-single-float (number)
+  (if (< number 0)
+      (complex (log (- number)) (coerce pi 'single-float))
+      (log number)))
+
+(defun log-double-float2 (number base)
+  (if (or (< number 0)
+          (< base 0))
+      (/ (log number) (log base))
+      (truly-the double-float (log number base))))
+
+(defun log-single-float2 (number base)
+  (if (or (< number 0)
+          (< base 0))
+      (/ (log number) (log base))
+      (truly-the single-float (log number base))))
 
 (declaim (start-block log))
 (defun log2/nonnegative-integer (x)
@@ -516,13 +510,10 @@
                                 log2e)
                              'single-float)))))
           (((foreach single-float double-float))
-           ;; Is (log -0) -infinity (libm.a) or -infinity + i*pi (Kahan)?
-           ;; Since this doesn't seem to be an implementation issue
-           ;; I (pw) take the Kahan result.
-           (if (float-sign-bit-set-p number)
+           ;; IEEE 754 says (log -0.0) should be -inf
+           (if (< number 0.0)
                (complex (log (- number)) (coerce pi '(dispatch-type number)))
-               (coerce (%log (coerce number 'double-float))
-                       '(dispatch-type number))))
+               (log number)))
           ((complex)
            (complex-log number))))))
 (declaim (end-block))
@@ -539,10 +530,8 @@
     (((foreach single-float double-float))
      (if (minusp number)
          (complex (coerce 0.0 '(dispatch-type number))
-                  (coerce (%sqrt (- (coerce number 'double-float)))
-                          '(dispatch-type number)))
-         (coerce (%sqrt (coerce number 'double-float))
-                 '(dispatch-type number))))
+                  (sqrt (- number)))
+          (sqrt number)))
     ((complex)
      (complex-sqrt number))))
 
@@ -561,9 +550,7 @@
          (rational
           (sqrt (+ (* rx rx) (* ix ix))))
          (single-float
-          (coerce (%hypot (coerce rx 'double-float)
-                          (coerce (truly-the single-float ix) 'double-float))
-                  'single-float))
+          (%hypotf rx ix))
          (double-float
           (%hypot rx (truly-the double-float ix))))))))
 
@@ -635,15 +622,13 @@
   (declare (explicit-check))
   (number-dispatch ((number number))
     ((rational)
-     (if (or (> number 1) (< number -1))
-         (complex-asin number)
-         (coerce (%asin (coerce number 'double-float)) 'single-float)))
+     (if (<= -1 number 1)
+         (%asinf (coerce number 'single-float))
+         (complex-asin number)))
     (((foreach single-float double-float))
-     (if (or (> number (coerce 1 '(dispatch-type number)))
-             (< number (coerce -1 '(dispatch-type number))))
-         (complex-asin (complex number))
-         (coerce (%asin (coerce number 'double-float))
-                 '(dispatch-type number))))
+     (if (<= (coerce -1 '(dispatch-type number)) number (coerce 1 '(dispatch-type number)))
+         (asin number)
+         (complex-asin (complex number))))
     ((complex)
      (complex-asin number))))
 
@@ -652,15 +637,13 @@
   (declare (explicit-check))
   (number-dispatch ((number number))
     ((rational)
-     (if (or (> number 1) (< number -1))
-         (complex-acos number)
-         (coerce (%acos (coerce number 'double-float)) 'single-float)))
+     (if (<= -1 number 1)
+         (%acosf (coerce number 'single-float))
+         (complex-acos number)))
     (((foreach single-float double-float))
-     (if (or (> number (coerce 1 '(dispatch-type number)))
-             (< number (coerce -1 '(dispatch-type number))))
-         (complex-acos (complex number))
-         (coerce (%acos (coerce number 'double-float))
-                 '(dispatch-type number))))
+     (if (<= (coerce -1 '(dispatch-type number)) number (coerce 1 '(dispatch-type number)))
+         (acos number)
+         (complex-acos (complex number))))
     ((complex)
      (complex-acos number))))
 
@@ -677,7 +660,17 @@
                            y
                            (float-sign y pi))
                        (float-sign y (sb-xc:/ pi 2)))
-                   (%atan2 y x))))
+                   (%atan2 y x)))
+             (atan2f (y x)
+               (declare (type single-float y x)
+                        (values single-float))
+               (if (zerop x)
+                   (if (zerop y)
+                       (if (not (float-sign-bit-set-p x))
+                           y
+                           (float-sign y (coerce pi 'single-float)))
+                       (float-sign y (sb-xc:/ (coerce pi 'single-float) 2)))
+                   (%atan2f y x))))
         (number-dispatch ((y real) (x real))
           ((double-float
             (foreach double-float single-float fixnum bignum ratio))
@@ -687,8 +680,7 @@
            (atan2 (coerce y 'double-float) x))
           (((foreach single-float fixnum bignum ratio)
             (foreach single-float fixnum bignum ratio))
-           (coerce (atan2 (coerce y 'double-float) (coerce x 'double-float))
-                   'single-float))))
+           (atan2f (coerce y 'single-float) (coerce x 'single-float)))))
       (number-dispatch ((y number))
         (handle-reals %atan y)
         ((complex)
@@ -746,12 +738,11 @@
      ;; acosh is complex if number < 1
      (if (< number 1)
          (complex-acosh number)
-         (coerce (%acosh (coerce number 'double-float)) 'single-float)))
+         (%acoshf (coerce number 'single-float))))
     (((foreach single-float double-float))
-     (if (< number (coerce 1 '(dispatch-type number)))
+     (if (< number 1)
          (complex-acosh (complex number))
-         (coerce (%acosh (coerce number 'double-float))
-                 '(dispatch-type number))))
+         (acosh number)))
     ((complex)
      (complex-acosh number))))
 
@@ -761,18 +752,15 @@
   (number-dispatch ((number number))
     ((rational)
      ;; atanh is complex if |number| > 1
-     (if (or (> number 1) (< number -1))
-         (complex-atanh number)
-         (coerce (%atanh (coerce number 'double-float)) 'single-float)))
+     (if (<= -1 number 1)
+         (%atanhf (coerce number 'single-float))
+         (complex-atanh number)))
     (((foreach single-float double-float))
-     (if (or (> number (coerce 1 '(dispatch-type number)))
-             (< number (coerce -1 '(dispatch-type number))))
-         (complex-atanh (complex number))
-         (coerce (%atanh (coerce number 'double-float))
-                 '(dispatch-type number))))
+     (if (<= -1 number 1)
+         (atanh number)
+         (complex-atanh (complex number))))
     ((complex)
      (complex-atanh number))))
-
 
 ;;;; not-OLD-SPECFUN stuff
 ;;;;
@@ -863,24 +851,6 @@
     ;; by one.
     (1- exponent)))
 
-;;; Compute an integer N such that 1 <= |2^N * x| < 2.
-;;; For the special cases, the following values are used:
-;;;    x             logb
-;;;   NaN            NaN
-;;;   +/- infinity   +infinity
-;;;   0              -infinity
-(defun logb (x)
-  (declare (type double-float x))
-  (cond ((float-nan-p x)
-         x)
-        ((float-infinity-p x) double-float-positive-infinity)
-        ((zerop x)
-         ;; The answer is negative infinity, but we are supposed to
-          ;; signal divide-by-zero, so do the actual division
-         (/ -1.0d0 x))
-        (t
-          (logb-finite x))))
-
 ;;; This function is used to create a complex number of the
 ;;; appropriate type:
 ;;;   Create complex number with real part X and imaginary part Y
@@ -955,7 +925,7 @@
   (declare (type (or complex rational) z))
   (multiple-value-bind (rho k)
       (cssqs z)
-    (declare (type (or (member 0d0) (double-float 0d0)) rho)
+    (declare (type (double-float 0d0) rho)
              (type fixnum k))
     (let ((x (float (realpart z) 1.0d0))
           (y (float (imagpart z) 1.0d0))
