@@ -149,7 +149,7 @@
 (declaim (inline whitespace[1]p whitespace[2]p))
 (declaim (inline constituentp terminating-macrop))
 (declaim (inline single-escape-p multiple-escape-p))
-(declaim (inline token-delimiterp))
+(declaim (inline token-delimiterp digit-char-p))
 
 ;;; the [1] and [2] here refer to ANSI glossary entries for "whitespace".
 (defun whitespace[1]p (char)
@@ -983,7 +983,9 @@ standard Lisp readtable when NIL."
 
 (defun read-right-paren (stream ignore)
   (declare (ignore ignore))
-  (simple-reader-error stream "unmatched close parenthesis"))
+  (with-simple-restart (continue "Ignore it")
+    (simple-reader-error stream "unmatched close parenthesis"))
+  (values))
 
 ;;; Read from the stream up to the next delimiter. Leave the resulting
 ;;; token in *READ-BUFFER*, and return three values:
@@ -2013,3 +2015,5 @@ extended <package-name>::<form-in-package> syntax."
     (cond ((not (whitespace[1]p char))
            (unread-char char stream)
            (return t)))))
+
+(declaim (maybe-inline digit-char-p))
