@@ -1006,7 +1006,7 @@
                                    :offset ,offset
                                    :to :result)
                                   ,name))
-                 *register-arg-names* *register-arg-offsets*))
+                 register-arg-names *register-arg-offsets*))
      ,@(when (eq return :fixed)
          '((:temporary (:scs (descriptor-reg) :from :eval) move-temp)
            (:temporary (:sc any-reg :from :eval :offset ocfp-offset) ocfp-temp)))
@@ -1025,14 +1025,18 @@
        (let* ((cur-nfp (current-nfp-tn vop))
               (filler
                 (remove nil
-                        (list ,@(if (eq return :tail)
-                                    '(:load-nargs
-                                      (when cur-nfp
-                                        :frob-nfp))
-                                    '(:load-nargs
-                                      (when cur-nfp
-                                        :frob-nfp)
-                                      :load-fp))))))
+                        (list ,@'(:load-nargs
+                                  (when cur-nfp
+                                    :frob-nfp)))))
+              ,@(unless (eq return :tail)
+                  `((new-fp-tn (cond ,@(and
+                                        (not variable)
+                                        '(((<= (if (consp nargs)
+                                                   (car nargs)
+                                                   nargs) register-arg-count)
+                                           csp-tn)))
+                                     (t
+                                      new-fp))))))
          (flet ((do-next-filler ()
                   (let* ((next (pop filler))
                          (what (if (consp next) (car next) next)))
@@ -1041,7 +1045,7 @@
                        ,@(if variable
                              `((inst sub nargs-pass csp-tn new-fp)
                                (inst asr nargs-pass nargs-pass (- word-shift n-fixnum-tag-bits))
-                               ,@(do ((arg *register-arg-names* (cddr arg))
+                               ,@(do ((arg register-arg-names (cddr arg))
                                       (i 0 (+ i 2))
                                       (insts))
                                      ((null arg) (nreverse insts))
@@ -1052,21 +1056,11 @@
                                (storew cfp-tn new-fp ocfp-save-offset))
                              '((unless (consp nargs)
                                  (load-immediate-word nargs-pass (fixnumize nargs))))))
-                      ,@(if (eq return :tail)
-                            '((:frob-nfp
-                               (inst add nsp-tn cur-nfp (add-sub-immediate
-                                                         (bytes-needed-for-non-descriptor-stack-frame)))))
-                            `((:frob-nfp
-                               (store-stack-tn nfp-save cur-nfp))
-                              (:load-fp
-                               (move cfp-tn (cond ,@(and
-                                                     (not variable)
-                                                     '(((<= (if (consp nargs)
-                                                                (car nargs)
-                                                                nargs) register-arg-count)
-                                                        csp-tn)))
-                                                  (t
-                                                   new-fp))))))
+                      (:frob-nfp
+                       ,(if (eq return :tail)
+                            `(inst add nsp-tn cur-nfp (add-sub-immediate
+                                                       (bytes-needed-for-non-descriptor-stack-frame)))
+                            `(store-stack-tn nfp-save cur-nfp)))
                       ((nil)))))
                 (insert-step-instrumenting ()
                   ;; Conditionally insert a conditional trap:
@@ -1107,10 +1101,6 @@
                      (do-next-filler)))
                   (insert-step-instrumenting)
                   (do-next-filler))))
-           (loop
-            (if filler
-                (do-next-filler)
-                (return)))
            ,@(case named
                ((t)
                 `((loadw lr name-pass fdefn-raw-addr-slot other-pointer-lowtag)
@@ -1120,16 +1110,24 @@
                 `((inst ldr lr (@ null-tn (load-store-offset (static-fun-offset fun))))
                   ,(if (eq return :tail)
                        `(inst add lr lr 4)))))
+           (loop
+            (if filler
+                (do-next-filler)
+                (return)))
+
 
            (note-this-location vop :call-site)
 
            ,(if named
                 (if (eq return :tail)
                     `(inst br lr)
-                    `(inst blr lr))
+                    `(progn
+                       ;; Load CFP just before calling for better profiling.
+                       (move cfp-tn new-fp-tn)
+                       (inst blr lr)))
                 (if (eq return :tail)
                     `(tail-call-unnamed lexenv lr fun-type)
-                    `(call-unnamed lexenv lr fun-type))))
+                    `(call-unnamed lexenv lr fun-type new-fp-tn))))
 
          ,@(ecase return
              (:fixed
@@ -1202,31 +1200,30 @@
     (t
      (assemble ()
        (when (eq type :designator)
+         (load-asm-routine lr 'call-symbol)
          (inst and tmp-tn lexenv lowtag-mask)
          (inst cmp tmp-tn fun-pointer-lowtag)
-         (inst b :eq call)
-         (invoke-asm-routine 'tail-call-symbol tmp-tn :tail t))
-       call
+         (inst b :ne call))
        (loadw lr lexenv closure-fun-slot fun-pointer-lowtag)
+       call
        (inst add lr lr 4)
        (inst br lr)))))
 
-(defun call-unnamed (lexenv lr type)
+(defun call-unnamed (lexenv lr type new-fp-tn)
   (case type
     (:symbol
-     (invoke-asm-routine 'call-symbol tmp-tn))
+     (invoke-asm-routine 'call-symbol lr :load-cfp new-fp-tn))
     (t
      (assemble ()
        (when (eq type :designator)
+         (load-asm-routine lr 'call-symbol)
          (inst and tmp-tn lexenv lowtag-mask)
          (inst cmp tmp-tn fun-pointer-lowtag)
-         (inst b :eq call)
-         (invoke-asm-routine 'call-symbol tmp-tn)
-         (inst b ret))
-       call
+         (inst b :ne call))
        (loadw lr lexenv closure-fun-slot fun-pointer-lowtag)
-       (inst blr lr)
-       ret))))
+       call
+       (move cfp-tn new-fp-tn)
+       (inst blr lr)))))
 
 ;;;; Unknown values return:
 
@@ -1247,11 +1244,7 @@
     ;; Interrupts leave two words of space for the new frame, so it's safe
     ;; to deallocate the frame before accessing OCFP/LR.
     (move csp-tn cfp-tn)
-    (loadw-pair cfp-tn ocfp-save-offset lr lra-save-offset cfp-tn)
-    ;; Clear the control stack, and restore the frame pointer.
-
-    ;; Out of here.
-    (lisp-return lr :single-value)))
+    (lisp-return lr :single-value t)))
 
 ;;; Do unknown-values return of a fixed number of values.  The Values are
 ;;; required to be set up in the standard passing locations.  Nvals is the
@@ -1289,15 +1282,11 @@
     (cond ((= nvals 1)
            ;; Clear the control stack, and restore the frame pointer.
            (move csp-tn cfp-tn)
-           (loadw-pair cfp-tn ocfp-save-offset lr lra-save-offset cfp-tn)
-           ;; Out of here.
-           (lisp-return lr :single-value))
+           (lisp-return lr :single-value t))
           (t
            ;; Establish the values pointer.
            (move val-ptr cfp-tn)
-           ;; restore the frame pointer and clear as much of the control
-           ;; stack as possible.
-           (loadw-pair cfp-tn ocfp-save-offset lr lra-save-offset cfp-tn)
+           ;; clear as much of the control stack as possible.
            (inst add csp-tn val-ptr (add-sub-immediate (* nvals n-word-bytes)))
            ;; Establish the values count.
            (load-immediate-word nargs (fixnumize nvals))
@@ -1306,7 +1295,7 @@
              (dolist (reg (subseq (list r0 r1 r2 r3) nvals))
                (move reg null-tn)))
            ;; And away we go.
-           (lisp-return lr :multiple-values)))))
+           (lisp-return lr :multiple-values t)))))
 
 ;;; Do unknown-values return of an arbitrary number of values (passed
 ;;; on the stack.)  We check for the common case of a single return
@@ -1319,14 +1308,13 @@
    (lra-arg)
    (vals-arg :scs (any-reg) :target vals)
    (nvals-arg :scs (any-reg) :target nvals))
-  (:temporary (:sc any-reg :offset nl2-offset :from (:argument 0)) old-fp)
   (:temporary (:sc any-reg :offset nl1-offset :from (:argument 2)) vals)
   (:temporary (:sc any-reg :offset nargs-offset :from (:argument 3)) nvals)
   (:temporary (:sc descriptor-reg :offset r0-offset) r0)
   (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
+  (:ignore old-fp-arg lra-arg)
   (:vop-var vop)
   (:generator 13
-    (maybe-load-stack-tn lr lra-arg)
     ;; Clear the number stack.
     (let ((cur-nfp (current-nfp-tn vop)))
       (when cur-nfp
@@ -1340,11 +1328,9 @@
     ;; Return with one value.
     (inst ldr r0 (@ vals-arg))
     (move csp-tn cfp-tn)
-    (move cfp-tn old-fp-arg)
-    (lisp-return lr :single-value)
+    (lisp-return lr :single-value t)
 
     NOT-SINGLE
-    (move old-fp old-fp-arg)
     (move vals vals-arg)
     (move nvals nvals-arg)
     (invoke-asm-routine 'return-multiple tmp-tn :tail t)))
