@@ -95,7 +95,7 @@
      (inst add ,lr ,lr 4)
      (inst br ,lr)))
 
-(defmacro lisp-return (lr return-style)
+(defmacro lisp-return (lr return-style &optional load-cfp-lr)
   "Return to RETURN-PC."
   `(progn
      ;; Indicate a single-valued return by clearing the Z flag
@@ -103,6 +103,8 @@
          (:single-value '((inst cmp null-tn 0)))
          (:multiple-values '((inst cmp zr-tn zr-tn)))
          (:known))
+     ,@(and load-cfp-lr
+            `((loadw-pair cfp-tn ocfp-save-offset ,lr lra-save-offset cfp-tn)))
      (inst ret ,lr)))
 
 ;;;; Stack TN's
@@ -143,6 +145,14 @@
 
 ;;;; Storage allocation:
 
+(defun generate-stack-overflow-check (vop size)
+  (let ((overflow (generate-error-code vop
+                                       'stack-allocated-object-overflows-stack-error
+                                       size)))
+    (loadw tmp-tn thread-tn thread-control-stack-end-slot)
+    (inst sub tmp-tn tmp-tn csp-tn)
+    (inst cmp tmp-tn size)
+    (inst b :le overflow)))
 
 ;;; This is the main mechanism for allocating memory in the lisp heap.
 ;;;
@@ -181,12 +191,6 @@
   (if stack-allocate-p
       (assemble ()
         (cond ((aligned-stack-p stack-allocate-p)
-               (assemble ()
-                 (inst tst csp-tn lowtag-mask)
-                 (inst b :eq skip)
-                 (inst add csp-tn csp-tn 8)
-                 (error-call nil 'sb-kernel::unreachable-error)
-                 skip)
                (move tmp-tn csp-tn)
                (inst add csp-tn csp-tn (add-sub-immediate size result-tn))
                (inst add result-tn tmp-tn lowtag))
@@ -344,7 +348,7 @@
          (progn
            (when ,sync
             (inst dmb :ishst))
-           (inst str (32-bit-reg zr-tn)
+           (inst str wzr-tn
                  (@ thread-tn
                     (* n-word-bytes thread-pseudo-atomic-bits-slot)))
            (inst ldr (32-bit-reg ,flag-tn)
@@ -425,7 +429,11 @@
         (:policy :fast-safe)
         (:args (object :scs (descriptor-reg))
                (index :scs (any-reg unsigned-reg signed-reg immediate))
-               (value :scs (,@scs zero)))
+               (value :scs (,@scs ,(case el-type
+                                     (double-float
+                                      '(double-immediate (eql (tn-value tn) 0d0)))
+                                     (t
+                                      'zero)))))
         (:arg-types ,type tagged-num ,el-type)
         (:vop-var vop)
         ,@(if barrierp '((:gc-barrier 0 2)
@@ -434,6 +442,10 @@
           ,@(when barrierp
               '((when barrier
                   (emit-gengc-barrier object nil tmp-tn t))))
+          ,@(case el-type
+              (double-float
+               '((when (sc-is value double-immediate)
+                   (setf value zr-tn)))))
           (sc-case index
             (immediate
              (inst str value (@ object (load-store-offset
@@ -488,55 +500,58 @@
 
 (defmacro define-partial-setter (name type size offset lowtag scs el-type
                                  &optional translate)
-  (let ((value `((value :scs ,scs
-                             :load-if (not (and (sc-is value immediate)
-                                                (eql (tn-value value) 0))))))
-        (setf-p (typep translate '(cons (eql setf)))))
-   `(define-vop (,name)
-      ,@(when translate
-          `((:translate ,translate)))
-      (:policy :fast-safe)
-      (:args ,@(when setf-p
-                 value)
-             (object :scs (descriptor-reg))
-             (index :scs (any-reg unsigned-reg signed-reg immediate))
-             ,@(unless setf-p
-                 value))
-      (:arg-types ,@(when setf-p
-                      `(,el-type))
-                  ,type
-                  tagged-num
-                  ,@(unless setf-p
-                      `(,el-type)))
-      (:generator 5
-        (when (sc-is value immediate)
-          (setf value zr-tn))
-        ,@(multiple-value-bind (op shift)
-              (ecase size
-                (:byte
-                 (values 'strb 0))
-                (:short
-                 (values 'strh 1))
-                ((:word :single-float)
-                 (values 'str 2)))
-            (let ((value (if (eq size :word)
-                             '(32-bit-reg value)
-                             'value)))
-              `((sc-case index
-                  (immediate
-                   (inst ,op ,value (@ object (load-store-offset
-                                               (+
-                                                (ash (tn-value index) ,shift)
-                                                (- (* ,offset n-word-bytes) ,lowtag))))))
-                  (t
-                   (let ((shift ,shift))
-                     (when (sc-is index any-reg)
-                       (decf shift n-fixnum-tag-bits))
-                     (inst add tmp-tn object (if (minusp shift)
-                                                 (asr index (- shift))
-                                                 (lsl index shift)))
-                     (inst ,op
-                           ,value (@ tmp-tn (- (* ,offset n-word-bytes) ,lowtag)))))))))))))
+  (multiple-value-bind (immediate-sc immediate-value)
+      (case el-type
+        (single-float (values 'single-immediate 0f0))
+        (t (values 'immediate 0)))
+    (let ((value `((value :scs (,@scs (,immediate-sc
+                                       (eql (tn-value tn) ,immediate-value))))))
+          (setf-p (typep translate '(cons (eql setf)))))
+      `(define-vop (,name)
+         ,@(when translate
+             `((:translate ,translate)))
+         (:policy :fast-safe)
+         (:args ,@(when setf-p
+                    value)
+                (object :scs (descriptor-reg))
+                (index :scs (any-reg unsigned-reg signed-reg immediate))
+                ,@(unless setf-p
+                    value))
+         (:arg-types ,@(when setf-p
+                         `(,el-type))
+                     ,type
+                     tagged-num
+                     ,@(unless setf-p
+                         `(,el-type)))
+         (:generator 5
+           (when (sc-is value ,immediate-sc)
+             (setf value wzr-tn))
+           ,@(multiple-value-bind (op shift)
+                 (ecase size
+                   (:byte
+                    (values 'strb 0))
+                   (:short
+                    (values 'strh 1))
+                   ((:word :single-float)
+                    (values 'str 2)))
+               (let ((value (if (eq size :word)
+                                '(32-bit-reg value)
+                                'value)))
+                 `((sc-case index
+                     (immediate
+                      (inst ,op ,value (@ object (load-store-offset
+                                                  (+
+                                                   (ash (tn-value index) ,shift)
+                                                   (- (* ,offset n-word-bytes) ,lowtag))))))
+                     (t
+                      (let ((shift ,shift))
+                        (when (sc-is index any-reg)
+                          (decf shift n-fixnum-tag-bits))
+                        (inst add tmp-tn object (if (minusp shift)
+                                                    (asr index (- shift))
+                                                    (lsl index shift)))
+                        (inst ,op
+                              ,value (@ tmp-tn (- (* ,offset n-word-bytes) ,lowtag))))))))))))))
 
 (defun load-inline-constant (dst &rest constant-descriptor)
   (inst load-from-label dst (cdr (apply #'register-inline-constant constant-descriptor))))
