@@ -285,10 +285,11 @@
 
 ;;; For each downward funarg, mark the funarg as dynamic extent. For
 ;;; now this only works on globally named functions.
-(defun dxify-downward-funargs (node dxable-args fun-name)
-  #+sb-xc-host
-  (declare (ignore fun-name))
-  (let (dynamic-extent)
+(defun dxify-downward-funargs (node)
+  (let* ((fun-name (combination-fun-source-name node nil))
+         (dxable-args (and fun-name
+                           (fun-name-dx-args fun-name)))
+         dynamic-extent)
     ;; Experience shows that users place incorrect DYNAMIC-EXTENT declarations
     ;; without due consideration and care. Since the declaration was ignored
     ;; in more contexts than not, it was relatively harmless.
@@ -299,96 +300,165 @@
     ;; that callers should always use more stack space. You should really
     ;; only do that if you don't also need an arbitrarily long call chain.
     ;; MAP and friends are good examples where this pertains]
-    (when #+sb-xc-host t                ; always trust our own code
-          #-sb-xc-host
-          (or (let ((pkg (sb-xc:symbol-package (fun-name-block-name fun-name))))
-                ;; callee "probably" won't get redefined
-                (or (not pkg)
-                    (package-locked-p pkg)
-                    (system-package-p pkg)
-                    (eq pkg *cl-package*)
-                    (basic-combination-fun-info node)))
-              (policy node (= safety 0)))
-          (dolist (arg-spec dxable-args)
-            (when (symbolp arg-spec)
-              ;; If there are keywords, we had better have a FUN-TYPE
-              (let ((fun-type (lvar-type (combination-fun node))))
-                ;; Can't do anything unless we can ascertain where
-                ;; the keyword arguments start.
-                (when (fun-type-p fun-type)
-                  (let* ((keys-index
-                           (+ (length (fun-type-required fun-type))
-                              (length (fun-type-optional fun-type))))
-                         (keywords-supplied
-                           (nthcdr keys-index (combination-args node))))
-                    ;; Everything in a keyword position needs to be
-                    ;; constant.
-                    (loop
-                      (unless (cdr keywords-supplied) (return))
-                      (let ((keyword (car keywords-supplied)))
-                        (unless (constant-lvar-p keyword)
-                          (return))
-                        (when (eq (lvar-value keyword) arg-spec)
-                          ;; Map it to a positional arg
-                          (setq arg-spec (1+ keys-index))
-                          (return))
-                        (setq keywords-supplied (cddr keywords-supplied))
-                        (incf keys-index 2)))))))
-            (when (integerp arg-spec)
-              (let* ((arg (or (nth arg-spec (combination-args node))
-                              (return-from dxify-downward-funargs)))
-                     (use (principal-lvar-use arg)))
-                (when (and (not (lvar-dynamic-extent arg))
-                           ;; We check that the use is a lambda so
-                           ;; that we don't end up getting notes about
-                           ;; not being able to allocate later.
-                           (ref-p use)
-                           (lambda-p (ref-leaf use))
-                           (not (leaf-dynamic-extent (functional-entry-fun (ref-leaf use)))))
-                  (let ((enclose (xep-enclose (ref-leaf use))))
-                    (cond ((and enclose
-                                (enclose-dynamic-extent enclose))
-                           (setf (leaf-dynamic-extent (functional-entry-fun (ref-leaf use)))
-                                 t))
-                          (t
-                           (unless dynamic-extent
-                             (setq dynamic-extent (insert-dynamic-extent node)))
-                           (setf (lvar-dynamic-extent arg) dynamic-extent)
-                           (push arg (dynamic-extent-values dynamic-extent))))))))))))
+    (when (and dxable-args
+               #-sb-xc-host                   ; always trust our own code
+               (or (let ((pkg (sb-xc:symbol-package
+                               (fun-name-block-name fun-name))))
+                     ;; callee "probably" won't get redefined
+                     (or (not pkg)
+                         (package-locked-p pkg)
+                         (system-package-p pkg)
+                         (eq pkg *cl-package*)
+                         (basic-combination-fun-info node)))
+                   (policy node (= safety 0))))
+      (dolist (arg-spec dxable-args)
+        (when (symbolp arg-spec)
+          ;; If there are keywords, we had better have a FUN-TYPE
+          (let ((fun-type (lvar-type (combination-fun node))))
+            ;; Can't do anything unless we can ascertain where
+            ;; the keyword arguments start.
+            (when (fun-type-p fun-type)
+              (let* ((keys-index
+                       (+ (length (fun-type-required fun-type))
+                          (length (fun-type-optional fun-type))))
+                     (keywords-supplied
+                       (nthcdr keys-index (combination-args node))))
+                ;; Everything in a keyword position needs to be
+                ;; constant.
+                (loop
+                  (unless (cdr keywords-supplied) (return))
+                  (let ((keyword (car keywords-supplied)))
+                    (unless (constant-lvar-p keyword)
+                      (return))
+                    (when (eq (lvar-value keyword) arg-spec)
+                      ;; Map it to a positional arg
+                      (setq arg-spec (1+ keys-index))
+                      (return))
+                    (setq keywords-supplied (cddr keywords-supplied))
+                    (incf keys-index 2)))))))
+        (when (integerp arg-spec)
+          (let* ((arg (or (nth arg-spec (combination-args node))
+                          (return-from dxify-downward-funargs)))
+                 (lvar (principal-lvar arg)))
+            (do-uses (use lvar)
+              (when (and (not (lvar-dynamic-extent arg))
+                         (ref-p use)
+                         (lambda-p (ref-leaf use))
+                         (not (leaf-dynamic-extent (functional-entry-fun (ref-leaf use))))
+                         ;; TODO: we need to do this because we don't
+                         ;; have enough smarts yet.
+                         (eq (node-home-lambda (xep-enclose (ref-leaf use)))
+                             (node-home-lambda node)))
+                (unless dynamic-extent
+                  (setq dynamic-extent (insert-dynamic-extent node)))
+                (setf (lvar-dynamic-extent arg) dynamic-extent)
+                (push arg (dynamic-extent-values dynamic-extent))))))))))
 
-;;; Make LVAR have DYNAMIC-EXTENT, recursively looking for otherwise
-;;; inaccessible potentially stack-allocatable parts. If LVAR already
-;;; has a different dynamic extent set, we don't do anything.
-(defun find-stack-allocatable-parts (lvar dynamic-extent)
+;;; Check that REF delivers a value to a combination which is DX safe
+;;; or whose result is that value and ends up being discarded.
+(defun ref-good-for-dx-p (ref)
+  (let* ((lvar (ref-lvar ref))
+         (dest (when lvar (lvar-dest lvar))))
+    (and (combination-p dest)
+         (case (combination-kind dest)
+           (:known
+            (awhen (combination-fun-info dest)
+              (or (ir1-attributep (fun-info-attributes it) dx-safe)
+                  (and (not (combination-lvar dest))
+                       (awhen (fun-info-result-arg it)
+                         (eql lvar (nth it (combination-args dest))))))))
+           (:local
+            (loop for arg in (combination-args dest)
+                  for var in (lambda-vars (combination-lambda dest))
+                  do (when (eq arg lvar)
+                       (return
+                         (dolist (ref (lambda-var-refs var) t)
+                           (unless (ref-good-for-dx-p ref)
+                             (return nil)))))
+                  finally (sb-impl::unreachable)))))))
+
+;;; Recursively look for otherwise inaccessible potentially
+;;; stack-allocatable parts in the uses of LVAR. If there is one,
+;;; bound LVAR's extent by DYNAMIC-EXTENT and return T. If LVAR
+;;; already has a different dynamic extent set, we don't do anything.
+(defun find-stack-allocatable-parts (lvar dynamic-extent &optional check-nesting)
   (declare (type lvar lvar)
            (type cdynamic-extent dynamic-extent))
-  (unless (lvar-dynamic-extent lvar)
-    (setf (lvar-dynamic-extent lvar) dynamic-extent))
-  (when (eq (lvar-dynamic-extent lvar) dynamic-extent)
+  (when (lvar-dynamic-extent lvar)
+    (aver (not (eq (lvar-dynamic-extent lvar) dynamic-extent)))
+    (return-from find-stack-allocatable-parts nil))
+  (let ((found-subpart-p nil))
     (do-uses (use lvar)
-      (when (use-good-for-dx-p use dynamic-extent)
-        (etypecase use
-          (cast
-           (find-stack-allocatable-parts (cast-value use) dynamic-extent))
-          (combination
-           ;; Don't propagate through &REST, for sanity.
-           (unless (eq (combination-fun-source-name use nil)
-                       '%listify-rest-args)
-             (dolist (arg (combination-args use))
-               (when (and arg
-                          (lvar-good-for-dx-p arg dynamic-extent))
-                 (find-stack-allocatable-parts arg dynamic-extent)))))
-          (ref
-           (let ((leaf (ref-leaf use)))
-             (typecase leaf
-               (lambda-var
-                (find-stack-allocatable-parts (let-var-initial-value leaf)
-                                              dynamic-extent))
-               (clambda
-                (let ((fun (functional-entry-fun leaf)))
-                  (setf (enclose-dynamic-extent (functional-enclose fun))
-                        dynamic-extent)
-                  (setf (leaf-dynamic-extent fun) t)))))))))))
+      (typecase use
+        (cast
+         (unless (cast-type-check use)
+           (when (find-stack-allocatable-parts (cast-value use) dynamic-extent)
+             (setq found-subpart-p t))))
+        (combination
+         (when (eq (combination-kind use) :known)
+           (let* ((info (combination-fun-info use))
+                  (stack-alloc-result (fun-info-stack-allocate-result info))
+                  (result-arg
+                    (let ((i (fun-info-result-arg info)))
+                      (and i (nth i (combination-args use))))))
+             (when (or (and result-arg
+                            (find-stack-allocatable-parts result-arg dynamic-extent))
+                       (and stack-alloc-result
+                            (funcall stack-alloc-result use)))
+               (setq found-subpart-p t)
+               (dolist (arg (combination-args use))
+                 (when (and arg (not (eq result-arg arg)))
+                   (find-stack-allocatable-parts arg dynamic-extent t)))))))
+        (ref
+         (let ((leaf (ref-leaf use)))
+
+           (typecase leaf
+             (lambda-var
+              ;; LET lambda var with no SETS.
+              (when (and (functional-kind-eq (lambda-var-home leaf) let)
+                         (not (lambda-var-sets leaf))
+                         (lexenv-contains-lambda (lambda-var-home leaf)
+                                                 (node-lexenv dynamic-extent))
+                         ;; Check the other refs are good.
+                         (dolist (ref (leaf-refs leaf) t)
+                           (unless (eq use ref)
+                             (when (not (ref-good-for-dx-p ref))
+                               (return nil)))))
+                (when (find-stack-allocatable-parts (let-var-initial-value leaf)
+                                                    dynamic-extent)
+                  (setq found-subpart-p t))))
+             (clambda
+              (when (functional-kind-eq leaf external)
+                (let* ((fun (functional-entry-fun leaf))
+                       (enclose (functional-enclose fun)))
+                  (when (and (or (not check-nesting)
+                                 ;; Allow (let ((x (lambda () v))) (let ((d x)) (dynamic-extent d)))
+                                 ;; but not (let ((x (lambda () v))) (let ((d (list x))) (dynamic-extent d)))
+                                 (lexenv-contains-lambda leaf (node-lexenv dynamic-extent)))
+                             (environment-closure (get-lambda-environment leaf))
+                             ;; To make sure the allocation is in the same
+                             ;; stack frame as the dynamic extent.
+                             (eq (node-home-lambda enclose)
+                                 (node-home-lambda dynamic-extent))
+                             ;; Check the other refs are good. At this
+                             ;; point, DXIFY-DOWNWARD-FUNARGS and
+                             ;; PROPAGATE-REF-DX should have marked
+                             ;; the p-lvar-ends of all good refs.
+                             (dolist (ref (leaf-refs leaf) t)
+                               (unless (eq use ref)
+                                 (multiple-value-bind (dest lvar)
+                                     (principal-lvar-end (node-lvar ref))
+                                   (declare (ignore dest))
+                                   (unless (lvar-dynamic-extent lvar)
+                                     (return nil))))))
+                    (unless (enclose-dynamic-extent enclose)
+                      (pushnew dynamic-extent
+                               (enclose-derived-dynamic-extents enclose)))
+                    (setf (leaf-dynamic-extent fun) t)
+                    (setq found-subpart-p t))))))))))
+    (when found-subpart-p
+      (setf (lvar-dynamic-extent lvar) dynamic-extent)
+      t)))
 
 ;;; Find all stack allocatable values in COMPONENT, setting
 ;;; appropriate dynamic extents for any lvar which may take on a stack
@@ -401,23 +471,31 @@
       (when (and (combination-p node)
                  (memq (basic-combination-kind node)
                        '(:full :unknown-keys :known)))
-        (let ((name (combination-fun-source-name node nil)))
-          (when name
-            (let ((dxable-args (fun-name-dx-args name)))
-              (when dxable-args
-                (dxify-downward-funargs node dxable-args name))))))))
+        (dxify-downward-funargs node))))
   ;; For each dynamic extent declared variable, each value that the
   ;; variable can take on is also dynamic extent.
   (dolist (lambda (component-lambdas component))
-    (dolist (var (lambda-vars lambda))
-      (when (leaf-dynamic-extent var)
-        (let ((values (mapcar #'set-value (basic-var-sets var))))
-          (when values
-            ;; This dynamic extent is over the whole environment and
-            ;; needs no cleanup code.
-            (let ((dynamic-extent (make-dynamic-extent values)))
-              (push dynamic-extent (lambda-dynamic-extents lambda))
+    (let* ((bind (lambda-bind lambda))
+           (lexenv (node-lexenv bind))
+           dynamic-extent)
+      (dolist (var (lambda-vars lambda))
+        (when (leaf-dynamic-extent var)
+          (let ((values (mapcar #'set-value (basic-var-sets var))))
+            (when values
+              ;; This dynamic extent is over the whole environment.
+              (unless dynamic-extent
+                (setf (node-lexenv bind) (make-lexenv :default lexenv))
+                (setq dynamic-extent
+                      (with-ir1-environment-from-node bind
+                        (make-dynamic-extent)))
+                (insert-node-after bind dynamic-extent)
+                (let ((cleanup (make-cleanup :dynamic-extent dynamic-extent)))
+                  (setf (dynamic-extent-cleanup dynamic-extent) cleanup)
+                  (aver (null (lexenv-cleanup lexenv)))
+                  (setf (lexenv-cleanup lexenv) cleanup))
+                (push dynamic-extent (lambda-dynamic-extents lambda)))
               (dolist (value values)
+                (push value (dynamic-extent-values dynamic-extent))
                 (setf (lvar-dynamic-extent value) dynamic-extent)))))))
     (dolist (let (lambda-lets lambda))
       (dolist (var (lambda-vars let))
@@ -441,28 +519,28 @@
     (dolist (dynamic-extent (lambda-dynamic-extents lambda))
       (dolist (lvar (dynamic-extent-values dynamic-extent))
         (aver (eq dynamic-extent (lvar-dynamic-extent lvar)))
-        ;; Check that the value hasn't been flushed somehow.
-        (when (lvar-uses lvar)
-          (cond ((lvar-good-for-dx-p lvar dynamic-extent)
-                 (find-stack-allocatable-parts lvar dynamic-extent)
-                 (unless (or (dynamic-extent-info dynamic-extent)
-                             (null (dynamic-extent-cleanup dynamic-extent)))
-                   (setf (dynamic-extent-info dynamic-extent) (make-lvar))))
-                (t
-                 (setf (lvar-dynamic-extent lvar) nil)))))))
-  (dolist (lambda (component-lambdas component))
-    (let ((fun (if (functional-kind-eq lambda optional)
-                   (lambda-optional-dispatch lambda)
-                   lambda)))
-      (when (leaf-dynamic-extent fun)
-        (let ((xep (functional-entry-fun fun)))
-          ;; We need to have a closure environment to dynamic-extent
-          ;; allocate.
-          (when (and xep (environment-closure (get-lambda-environment xep)))
-            (let ((dynamic-extent
-                    (enclose-dynamic-extent (functional-enclose fun))))
-              (unless (dynamic-extent-info dynamic-extent)
-                (setf (dynamic-extent-info dynamic-extent) (make-lvar)))))))))
+        (setf (lvar-dynamic-extent lvar) nil)
+        (when (find-stack-allocatable-parts lvar dynamic-extent)
+          (setf (dynamic-extent-info dynamic-extent) (make-lvar))))))
+  (dolist (xep (component-lambdas component))
+    (when (and (functional-kind-eq xep external)
+               (leaf-dynamic-extent (functional-entry-fun xep))
+               ;; We need to have a closure environment to
+               ;; stack allocate.
+               (environment-closure (get-lambda-environment xep)))
+      (let* ((enclose (xep-enclose xep))
+             (dynamic-extent (enclose-dynamic-extent enclose))
+             (derived-dynamic-extents
+               (enclose-derived-dynamic-extents enclose)))
+        (cond (dynamic-extent
+               (aver (null derived-dynamic-extents))
+               (unless (dynamic-extent-info dynamic-extent)
+                 (setf (dynamic-extent-info dynamic-extent) (make-lvar))))
+              (derived-dynamic-extents
+               (aver (null (enclose-dynamic-extent enclose)))
+               (let ((lvar (make-lvar)))
+                 (dolist (dynamic-extent derived-dynamic-extents)
+                   (setf (dynamic-extent-info dynamic-extent) lvar))))))))
   (values))
 
 ;;;; cleanup emission
@@ -582,8 +660,8 @@
             (when (and (immediately-used-p result use)
                        (not (and (combination-p use)
                                  (lvar-fun-is (combination-fun use) '(break))))
+                       (basic-combination-p use)
                        (or (not (eq (node-derived-type use) *empty-type*))
-                           (not (basic-combination-p use))
                            ;; This prevents external entry points from
                            ;; showing up in the backtrace: we always
                            ;; want tail calls inside XEPs to the
