@@ -604,7 +604,8 @@
                       reference-constant
                       expand-compiler-macro
                       maybe-reanalyze-functional
-                      ir1-convert-common-functoid))
+                      ir1-convert-common-functoid
+                      record-macroexpand-source-path))
 
 ;;; Translate FORM into IR1. The code is inserted as the NEXT of the
 ;;; CTRAN START. RESULT is the LVAR which receives the value of the
@@ -889,6 +890,26 @@
                           (find-free-fun fun "shouldn't happen! (no-cmacro)")
                           form))))
 
+(defvar *equal-source-paths*)
+
+;;; Some macros call macroexpand-1 and then copy its results.
+(defun record-macroexpand-source-path (original-form expanded env)
+  (when (boundp '*equal-source-paths*)
+    (let ((env (if (and (boundp '*lexenv*)
+                        (or (not (lexenv-p env))
+                            (null-lexenv-p env)))
+                   *lexenv*
+                   env)))
+      (when (and (lexenv-p env)
+                 (policy env (> debug 1)))
+        (unless *equal-source-paths*
+          (setf *equal-source-paths* (make-hash-table :test #'equal)))
+        (let ((path (get-source-path original-form)))
+          (when path
+            (push path (gethash expanded *equal-source-paths*))
+            (let ((*lexenv* env))
+              (recover-source-paths original-form expanded))))))))
+
 ;;; This may produce false matches when there are multiple copies and
 ;;; they are reordered, duplicated or omitted. Still better than
 ;;; nothing.
@@ -914,7 +935,9 @@
         (let ((seen (alloc-xset)))
           (labels ((rec (form)
                      (unless (xset-member-p form seen)
-                       (let ((original (gethash form equal-table)))
+                       (let ((original (or (gethash form equal-table)
+                                           (and *equal-source-paths*
+                                                (gethash form *equal-source-paths*)))))
                          (when original
                            (let ((location
                                    (if (cdr original)
@@ -963,7 +986,8 @@
                          (t
                           (compiler-error "~@<~A~@:_ ~A~:>"
                                           (wherestring) c))))))
-      (let ((result (funcall (valid-macroexpand-hook) fun form *lexenv*)))
+      (let* (*equal-source-paths*
+             (result (funcall (valid-macroexpand-hook) fun form *lexenv*)))
         #-sb-xc-host
         (recover-source-paths form result)
         result))))
@@ -1273,7 +1297,6 @@
                                     (list old-type type var-name))))
                                 (bound-var
                                  (setf (leaf-type bound-var) int
-                                       (leaf-defined-type bound-var) int
                                        (leaf-where-from bound-var) :declared))
                                 (t
                                  (restr (cons var int)))))))
@@ -1412,6 +1435,15 @@
         (t
          (setf (lambda-var-constant var) t))))))
 
+(defun process-no-debug-decl (spec vars)
+  (dolist (name (rest spec))
+    (let ((var (find-in-bindings vars name)))
+      (cond
+        ((not var)
+         (style-warn "No ~s variable" name))
+        (t
+         (setf (lambda-var-no-debug var) t))))))
+
 ;;; Return a DEFINED-FUN which copies a GLOBAL-VAR but for its INLINEP
 ;;; (and TYPE if notinline), plus type-restrictions from the lexenv.
 (defun make-new-inlinep (var inlinep local-type)
@@ -1542,11 +1574,6 @@
          (setf (lambda-var-ignorep var) t)))))
   (values))
 
-(defvar *stack-allocate-dynamic-extent* t
-  "If true (the default), the compiler believes DYNAMIC-EXTENT declarations
-and stack allocates otherwise inaccessible parts of the object whenever
-possible.")
-
 (defun process-dynamic-extent-decl (names vars fvars)
   (if *stack-allocate-dynamic-extent*
       (dolist (name names)
@@ -1681,6 +1708,9 @@ possible.")
        (constant-value
         (process-constant-decl spec vars)
         res)
+       (no-debug
+        (process-no-debug-decl spec vars)
+        res)
        ;; We may want to detect LAMBDA-LIST and VALUES decls here,
        ;; and report them as "Misplaced" rather than "Unrecognized".
        (t
@@ -1882,7 +1912,7 @@ possible.")
   (declare (list names))
   (let ((type (specifier-type spec)))
     (unless (csubtypep type (specifier-type 'function))
-      (error "Not a function type: ~/sb-impl:print-type/" spec))
+      (error "Not a function type: ~/sb-impl:print-type-specifier/" spec))
     (dolist (name names)
       (process-1-ftype-proclamation name type))))
 
