@@ -186,12 +186,20 @@
 
 (defun unboxed-specialized-return-p (name)
   (and (typep name '(cons (eql sb-impl::specialized-xep)))
-       (let ((type (fun-type-returns (specifier-type `(function ,@(cddr name))))))
+       (let ((type (values-specifier-type (cadddr name))))
          (and (values-type-p type)
               (not (or (values-type-optional type)
                        (values-type-rest type)))
               type))))
 
+(defun unboxed-return-p (combination)
+  (let ((fun (basic-combination-fun combination)))
+    (or (let ((info (basic-combination-fun-info combination)))
+          (and info
+               (ir1-attributep (fun-info-attributes info) unboxed-return)
+               (fun-type-returns (lvar-type fun))))
+        (unboxed-specialized-return-p (lvar-fun-name fun)))))
+
 ;;; If TAIL-P is true, then we check to see whether the call can
 ;;; really be a tail call by seeing if this function's return
 ;;; convention is :UNKNOWN. If so, we move the call block successor
@@ -202,10 +210,7 @@
   (declare (type basic-combination call))
   (let ((tails (and (node-tail-p call)
                     (lambda-tail-set (node-home-lambda call))))
-        (unboxed-return (or (let ((info (basic-combination-fun-info call)))
-                              (and info
-                                   (ir1-attributep (fun-info-attributes info) unboxed-return)))
-                            (unboxed-specialized-return-p (lvar-fun-name (basic-combination-fun call))))))
+        (unboxed-return (unboxed-return-p call)))
     (cond ((not tails))
           ((eq (return-info-kind (tail-set-info tails))
                (if unboxed-return
@@ -236,6 +241,8 @@
   (let ((kind (basic-combination-kind call))
         (info (basic-combination-fun-info call)))
 
+    (rewrite-full-call call)
+
     (dolist (arg (basic-combination-args call))
       (unless (lvar-info arg)
         (setf (lvar-info arg)
@@ -253,8 +260,7 @@
          (when (basic-combination-info call)
            (signal-delayed-combination-condition call))
          (setf (basic-combination-kind call) :full))
-       (setf (basic-combination-info call) :full)
-       (rewrite-full-call call))))
+       (setf (basic-combination-info call) :full))))
   (annotate-fun-lvar (basic-combination-fun call))
   (values))
 
@@ -467,6 +473,10 @@
   (annotate-ordinary-lvar (set-value node))
   (values))
 
+(defoptimizer (%%primitive ltn-annotate) ((template &rest args) node)
+  (ltn-default-call node)
+  (setf (basic-combination-info node) (lvar-value template)))
+
 ;;; If the only use of the TEST lvar is a combination annotated with a
 ;;; conditional template, then don't annotate the lvar so that IR2
 ;;; conversion knows not to emit any code, otherwise annotate as an
@@ -512,7 +522,7 @@
   (declare (type cdynamic-extent node))
   (let ((lvar (dynamic-extent-info node))
         (2comp (component-info (node-component node))))
-    (when lvar
+    (when (and lvar (not (lvar-info lvar)))
       (setf (ir2-component-stack-allocates-p 2comp) t)
       (setf (lvar-dest lvar) node)
       (let ((info (make-ir2-lvar *backend-t-primitive-type*)))
