@@ -33,8 +33,8 @@
     (let ((name (uncross name)))
       (setq *undefined-warnings*
             (delete-if (lambda (x)
-                         (and (equal (undefined-warning-name x) name)
-                              (eq (undefined-warning-kind x) kind)))
+                         (and (eq (undefined-warning-kind x) kind)
+                              (equal (undefined-warning-name x) name)))
                        *undefined-warnings*))))
   (values))
 
@@ -159,6 +159,8 @@
 ;;; the default.
 (defun undefine-fun-name (name)
   (when name
+    #-sb-xc-host
+    (sb-impl::remove-specialized-xep name)
     (macrolet ((frob (&rest types)
                  `(clear-info-values
                    name ',(mapcar (lambda (x)
@@ -457,9 +459,6 @@
             software version name replacement))
          (setf (info :variable :deprecated name) info))
         (type
-         (when (eq state :final)
-           (sb-impl::setup-type-in-final-deprecation
-            software version name replacement))
          (setf (info :type :deprecated name) info))))))
 
 (defun process-declaration-declaration (name form)
