@@ -221,7 +221,7 @@ int thread_wait_until_not(int undesired_state,
  * so this works out fine. The pthread ID is not an acceptable substitute.
  * (#+win32 can have WaitOnAddress use an 8-byte value, but we don't)
  */
-static int get_nonzero_tid()
+int sb_get_os_thread_id()
 {
     int tid = sb_GetTID();
 #ifdef LISP_FEATURE_SB_FUTEX
@@ -364,10 +364,24 @@ static lispobj cons_lisp_thread(struct thread* thread)
 #ifndef LISP_FEATURE_64_BIT
     instance->sw_observed_internal_real_time_delta_millisec = 0x7FFFFFFF;
     instance->internal_real_time = NIL;
+#endif
+    instance->alloc_histogram = NIL;
+#ifdef LISP_FEATURE_X86_64
+    struct vector* v = (void*)gc_general_alloc(r, (2+ALLOC_HISTOGRAM_WORDS)*N_WORD_BYTES,
+                                               PAGE_TYPE_MIXED);
+    v->header = SIMPLE_ARRAY_UNSIGNED_BYTE_64_WIDETAG;
+    v->length_ = make_fixnum(ALLOC_HISTOGRAM_WORDS);
+    instance->alloc_histogram = make_lispobj(v, OTHER_POINTER_LOWTAG);
 #endif
     return make_lispobj(instance, INSTANCE_POINTER_LOWTAG);
 }
 
+#ifdef LISP_FEATURE_64_BIT
+#  define SET_LISPTHREAD_TID(x) x->os_tid = make_fixnum(sb_get_os_thread_id())
+#else // it's a raw slot
+#  define SET_LISPTHREAD_TID(x) x->uw_os_tid = sb_get_os_thread_id()
+#endif
+
 void create_main_lisp_thread(lispobj function) {
 #ifdef LISP_FEATURE_WIN32
     InitializeCriticalSection(&all_threads_lock);
@@ -377,16 +391,14 @@ void create_main_lisp_thread(lispobj function) {
     struct thread *th = alloc_thread_struct(0);
     if (!th || arch_os_thread_init(th)==0 || !init_shared_attr_object())
         lose("can't create initial thread");
-    th->state_word.sprof_enable = 1;
+    th->sprof_enable = make_fixnum(1);
 #if defined LISP_FEATURE_SB_THREAD && !defined LISP_FEATURE_GCC_TLS && !defined LISP_FEATURE_WIN32
     pthread_key_create(&current_thread, 0);
 #endif
 #if defined LISP_FEATURE_DARWIN && defined LISP_FEATURE_SB_THREAD
     pthread_key_create(&ignore_stop_for_gc, 0);
 #endif
-#if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
-    __attribute__((unused)) lispobj *args = NULL;
-#endif
+
     ASSOCIATE_OS_THREAD(th);
     ASSIGN_CURRENT_THREAD(th);
 #if THREADS_USING_GCSIGNAL && \
@@ -395,27 +407,24 @@ void create_main_lisp_thread(lispobj function) {
     unblock_gc_stop_signal();
 #endif
     link_thread(th);
-    th->os_kernel_tid = get_nonzero_tid();
     th->lisp_thread = cons_lisp_thread(th);
+    SET_LISPTHREAD_TID(LISPTHREAD(th));
 
 #ifndef LISP_FEATURE_WIN32
-    protect_control_stack_hard_guard_page(1, NULL);
+    protect_control_stack_hard_guard_page(1, th);
 #endif
-    protect_binding_stack_hard_guard_page(1, NULL);
-    protect_alien_stack_hard_guard_page(1, NULL);
+    protect_binding_stack_hard_guard_page(1, th);
+    protect_alien_stack_hard_guard_page(1, th);
 #ifndef LISP_FEATURE_WIN32
-    protect_control_stack_guard_page(1, NULL);
+    protect_control_stack_guard_page(1, th);
 #endif
-    protect_binding_stack_guard_page(1, NULL);
-    protect_alien_stack_guard_page(1, NULL);
+    protect_binding_stack_guard_page(1, th);
+    protect_alien_stack_guard_page(1, th);
 
 #if defined(LISP_FEATURE_DARWIN) && defined(LISP_FEATURE_X86_64)
     set_thread_stack(th->control_stack_end);
 #endif
 
-#ifdef LISP_FEATURE_PPC64
-    NIL_SYMBOL_SLOTS_START[-1] = (uword_t)gc_card_mark;
-#endif
 #ifdef COLLECT_GC_STATS
     atexit(summarize_gc_stats);
 #endif
@@ -424,7 +433,7 @@ void create_main_lisp_thread(lispobj function) {
      * of the current stack */
 #if !(defined(LISP_FEATURE_WIN32) && !defined(OS_THREAD_STACK)) \
     && (defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
-    call_into_lisp_first_time(function,args,0);
+    call_into_lisp_first_time(function, NULL, 0);
 #else
     funcall0(function);
 #endif
@@ -434,7 +443,9 @@ void create_main_lisp_thread(lispobj function) {
 
 void sb_posix_after_fork() { // for use by sb-posix:fork
     struct thread* th = get_sb_vm_thread();
-    th->os_kernel_tid = get_nonzero_tid();
+    // There's no reason for a GC to occur, so this use of th->lisp_thread
+    // is OK for precise GC.
+    SET_LISPTHREAD_TID(LISPTHREAD(th));
 #ifdef LISP_FEATURE_DARWIN
     extern void darwin_reinit();
     darwin_reinit();
@@ -452,6 +463,7 @@ init_new_thread(struct thread *th,
                 init_thread_data __attribute__((unused)) *scribble,
                 int guardp)
 {
+
     ASSIGN_CURRENT_THREAD(th);
     if(arch_os_thread_init(th)==0) {
         /* FIXME: handle error */
@@ -464,18 +476,18 @@ init_new_thread(struct thread *th,
 
 #ifndef LISP_FEATURE_WIN32
     if (guardp & GUARD_CONTROL_STACK)
-        protect_control_stack_guard_page(1, NULL);
+        protect_control_stack_guard_page(1, th);
 #endif
     if (guardp & GUARD_BINDING_STACK)
-        protect_binding_stack_guard_page(1, NULL);
+        protect_binding_stack_guard_page(1, th);
     if (guardp & GUARD_ALIEN_STACK)
-        protect_alien_stack_guard_page(1, NULL);
+        protect_alien_stack_guard_page(1, th);
 
     /* Since GC can only know about this thread from the all_threads
      * list and we're just adding this thread to it, there is no
      * danger of deadlocking even with SIG_STOP_FOR_GC blocked (which
      * it is not). */
-#ifdef LISP_FEATURE_SB_SAFEPOINT
+#if defined(LISP_FEATURE_SB_SAFEPOINT) || defined(LISP_FEATURE_NONSTOP_FOREIGN_CALL)
     csp_around_foreign_call(th) = (lispobj)scribble;
 #endif
     __attribute__((unused)) int lock_ret = mutex_acquire(&all_threads_lock);
@@ -494,8 +506,6 @@ init_new_thread(struct thread *th,
 #endif
 }
 
-lispobj remset_transfer_list;
-
 static void
 unregister_thread(struct thread *th,
                   init_thread_data __attribute__((unused)) *scribble)
@@ -608,6 +618,7 @@ void* new_thread_trampoline(void* arg)
     // due to the pinning via *STARTING-THREADS*.
     struct thread_instance *lispthread = (void*)native_pointer(th->lisp_thread);
     if (lispthread->ephemeral_p == LISP_T) th->state_word.user_thread_p = 0;
+    SET_LISPTHREAD_TID(lispthread);
 
     struct vector* startup_info = VECTOR(lispthread->startup_info); // 'lispthread' is pinned
     gc_assert(header_widetag(startup_info->header) == SIMPLE_VECTOR_WIDETAG);
@@ -637,7 +648,6 @@ void* new_thread_trampoline(void* arg)
     && !defined LISP_FEATURE_SB_SAFEPOINT
     th->control_stack_end = (lispobj*)&arg + 1;
 #endif
-    th->os_kernel_tid = get_nonzero_tid();
     init_new_thread(th, SCRIBBLE, 0);
     // Passing the untagged pointer ensures 2 things:
     // - that the pinning mechanism works as designed, and not just by accident.
@@ -746,7 +756,6 @@ static void attach_os_thread(init_thread_data *scribble)
     unblock_gc_stop_signal();
 #endif
 
-    th->os_kernel_tid = get_nonzero_tid();
     /* win32: While ASSOCIATE_OS_THREAD performs a relatively expensive DuplicateHandle(),
      * simplicity here is preferable to the complexity entailed by memoizing the handle
      * in a TLS slot and registering a waiter on the foreign thread to close to handle.
@@ -915,7 +924,6 @@ uword_t create_lisp_thread(struct thread* th)
                      CREATE_SUSPENDED, &tid);
     bool success = th->os_thread != 0;
     if (success) {
-        th->os_kernel_tid = tid;
         ResumeThread((HANDLE)th->os_thread);
     }
     return success;
@@ -1013,24 +1021,56 @@ void gc_stop_the_world()
         if (th != me) {
             gc_assert(th->os_thread != 0);
             struct extra_thread_data *semaphores = thread_extra_data(th);
+            bool foreign = 0;
+#ifdef LISP_FEATURE_NONSTOP_FOREIGN_CALL
+            pthread_mutex_t *exit_lock = &semaphores->foreign_exit_lock;
+            gc_assert(mutex_acquire(exit_lock));
+            lispobj csp = set_thread_foreign_call_trigger(th, 0);
+            foreign = csp != 0;
+#endif
+
             os_sem_wait(&semaphores->state_sem);
             int state = get_thread_state(th);
             if (state == STATE_RUNNING) {
-                rc = pthread_kill(th->os_thread,SIG_STOP_FOR_GC);
-                /* This used to bogusly check for ESRCH.
-                 * I changed the ESRCH case to just fall into lose() */
-                if (rc) lose("cannot suspend thread %p: %d, %s",
-                     // KLUDGE: assume that os_thread can be cast as pointer.
-                     // See comment in 'interr.h' about that.
-                     (void*)th->os_thread, rc, strerror(rc));
+                if (!foreign) {
+                    rc = pthread_kill(th->os_thread,SIG_STOP_FOR_GC);
+                    /* This used to bogusly check for ESRCH.
+                     * I changed the ESRCH case to just fall into lose() */
+                    if (rc) lose("cannot suspend thread %p: %d, %s",
+                                 // KLUDGE: assume that os_thread can be cast as pointer.
+                                 // See comment in 'interr.h' about that.
+                                 (void*)th->os_thread, rc, strerror(rc));
+
+                }
+#ifdef LISP_FEATURE_NONSTOP_FOREIGN_CALL
+                else {
+                    if (read_TLS(GC_INHIBIT,th) != NIL) {
+                        /* It's a foreign call but inside without-gcing,
+                           handle_foreign_call_trigger will set
+                           *stop-for-gc-pending*, need to unlock the exit
+                           lock to reach the end of without-gcing* */
+                        semaphores->gc_inhibited = 2;
+                        gc_assert(mutex_release(exit_lock));
+                    } else {
+#ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
+                        th->control_stack_pointer = (lispobj*)csp;
+#endif
+                        semaphores->gc_inhibited = 1;
+                    }
+                }
+#endif
             }
             os_sem_post(&semaphores->state_sem);
         }
     }
     for_each_thread(th) {
         if (th != me) {
-            __attribute__((unused)) int state = thread_wait_until_not(STATE_RUNNING, th);
-            gc_assert(state != STATE_RUNNING);
+#ifdef LISP_FEATURE_NONSTOP_FOREIGN_CALL
+            if (thread_extra_data(th)->gc_inhibited != 1)
+#endif
+            {
+                thread_wait_until_not(STATE_RUNNING, th);
+            }
         }
     }
     event0("/gc_stop_the_world:end");
@@ -1064,14 +1104,32 @@ void gc_start_the_world()
     for_each_thread(th) {
         gc_assert(th->os_thread);
         if (th != me) {
-          /* I don't know if a normal load is fine here. I think we can't read
-           * any value other than what was already observed?
-           * No harm in being cautious though with regard to compiler reordering */
+            /* I don't know if a normal load is fine here. I think we can't read
+             * any value other than what was already observed?
+             * No harm in being cautious though with regard to compiler reordering */
             int state = get_thread_state(th);
-            if (state != STATE_DEAD) {
-                if(state != STATE_STOPPED)
-                    lose("gc_start_the_world: bad thread state %x", state);
-                set_thread_state(th, STATE_RUNNING, 0);
+#ifdef LISP_FEATURE_NONSTOP_FOREIGN_CALL
+            bool foreign = csp_around_foreign_call(th);
+            set_thread_foreign_call_trigger(th, 1);
+
+            if (thread_extra_data(th)->gc_inhibited != 2) // already released
+                gc_assert(mutex_release(&thread_extra_data(th)->foreign_exit_lock));
+
+            thread_extra_data(th)->gc_inhibited = 0;
+
+            if (foreign) {
+#ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
+                th->control_stack_pointer = 0;
+#endif
+            }
+            else
+#endif
+            {
+                if (state != STATE_DEAD) {
+                    if (state != STATE_STOPPED)
+                        lose("gc_start_the_world: bad thread state %x", state);
+                    set_thread_state(th, STATE_RUNNING, 0);
+                }
             }
         }
     }
