@@ -48,7 +48,44 @@
 #include "validate.h"
 
 int arch_os_thread_init(struct thread *thread) {
+#ifdef MEMORY_SANITIZER
+    extern __thread unsigned long __msan_param_tls[];
+    thread->msan_param_tls = (uword_t)&__msan_param_tls[0];
+#endif
     stack_t sigstack;
+#ifdef ADDRESS_SANITIZER // maybe skip setting up our own altstack
+    /* Asan seems to think it should call munmap() on the alternate signal stack because
+     * it believes that it - and only it - called sigaltstack to begin with, due to:
+     *   COMMON_FLAG(bool, use_sigaltstack, true,
+     *               "If set, uses alternate stack for signal handling.")
+     * in llvm-project/compiler-rt/lib/sanitizer_common/sanitizer_flags.inc
+     * and
+     *   if (common_flags()->use_sigaltstack) SetAlternateSignalStack();
+     * which picks a kernel-assigned address. This ocurs before we get here.
+     *
+     * I suppose when running SBCL, you're supposed to pass a sanitizer flag,
+     * but how would users know? Our choices are either to avoid using our own altstack,
+     * or undo our altstack and put it back to the ASan-specified stack just before the
+     * thread is joined. The consequence of doing otherwise is to see the sanitizer
+     * trip over its own shoelaces after we've freed a thread struct, then it queries
+     * the OS for the altstack and frees that, which is a double-free, leading to:
+#0  __sanitizer::ReportMunmapFailureAndDie () at third_party/llvm/llvm-project/compiler-rt/lib/sanitizer_common/sanitizer_common.cpp:67
+#1  0x0000000000364b3d in __sanitizer::UnmapOrDie () at third_party/llvm/llvm-project/compiler-rt/lib/sanitizer_common/sanitizer_posix.cpp:62
+#2  0x00000000003662df in __sanitizer::UnsetAlternateSignalStack () at third_party/llvm/llvm-project/compiler-rt/lib/sanitizer_common/sanitizer_posix_libcdep.cpp:210
+#3  0x000000000034aceb in __asan::AsanThread::Destroy () at third_party/llvm/llvm-project/compiler-rt/lib/asan/asan_thread.cpp:135
+#4  0x00007f33072f0271 in __GI___nptl_deallocate_tsd () at ./nptl/nptl_deallocate_tsd.c:73
+#5  __GI___nptl_deallocate_tsd () at ./nptl/nptl_deallocate_tsd.c:22
+#6  0x00007f33072f2a4e in start_thread (arg=<optimized out>) at ./nptl/pthread_create.c:456
+#7  0x00007f33073707b8 in __GI___clone3 () at ../sysdeps/unix/sysv/linux/x86_64/clone3.S:78
+     *
+     * In trying to track this down, I did first attempt to completely disable signal stacks,
+     * and then due to another bug I caused (in trying to use FORMAT after streams were closed)
+     * I got a stack overflow, and the sanitizer didn't help.*/
+    memset(&sigstack, 0, sizeof sigstack);
+    int res = sigaltstack(0, &sigstack);
+    if (res == 0 && sigstack.ss_sp) return 1;
+#endif
+    // why is this guard even here? It's always C_STACK_IS_CONTROL_STACK for x86-64
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
     /* Signal handlers are run on the control stack, so if it is exhausted
      * we had better use an alternate stack for whatever signal tells us
@@ -59,10 +96,6 @@ int arch_os_thread_init(struct thread *thread) {
     if(sigaltstack(&sigstack,0)<0) {
         lose("Cannot sigaltstack: %s",strerror(errno));
     }
-#endif
-#ifdef MEMORY_SANITIZER
-    extern __thread unsigned long __msan_param_tls[];
-    ((lispobj*)thread)[THREAD_MSAN_PARAM_TLS_SLOT] = (uword_t)&__msan_param_tls[0];
 #endif
     return 1;
 }
@@ -75,6 +108,8 @@ int arch_os_thread_cleanup(struct thread __attribute__((unused)) *thread) {
     return 1;
 }
 
+/* Visit all registers even if lisp doesn't store pointers there, C
+code might be using any register for lisp pointers. */
 void visit_context_registers(void (*p)(os_context_register_t,void*),
                              os_context_t *context, void* arg)
 {
@@ -83,13 +118,7 @@ void visit_context_registers(void (*p)(os_context_register_t,void*),
     // This is the order the registers appear in gregset_t (which makes no difference of course).
     // Not sure why the order is so kooky.
     p(m->gregs[REG_R8 ], arg); p(m->gregs[REG_R9 ], arg); p(m->gregs[REG_R10], arg);
-    p(m->gregs[REG_R11], arg);
-#ifndef LISP_FEATURE_SOFT_CARD_MARKS
-    p(m->gregs[REG_R12], arg);  /* CARD_TABLE_REG */
-#endif
-#ifndef LISP_FEATURE_SB_THREAD
-    p(m->gregs[REG_R13], arg);  /* THREAD_BASE_REG */
-#endif
+    p(m->gregs[REG_R11], arg); p(m->gregs[REG_R12], arg); p(m->gregs[REG_R13], arg);
     p(m->gregs[REG_R14], arg); p(m->gregs[REG_R15], arg);
     p(m->gregs[REG_RDI], arg); p(m->gregs[REG_RSI], arg); p(m->gregs[REG_RBX], arg);
     p(m->gregs[REG_RDX], arg); p(m->gregs[REG_RAX], arg); p(m->gregs[REG_RCX], arg);
@@ -175,11 +204,16 @@ os_context_fp_addr(os_context_t *context)
     return (os_context_register_t*)&context->uc_mcontext.gregs[REG_RBP];
 }
 
-unsigned long
+unsigned int
 os_context_fp_control(os_context_t *context)
 {
-    return (uintptr_t)&context->uc_mcontext.gregs[REG_RSP];
+    return context->uc_mcontext.fpregs->mxcsr ^ (0x3F << 7);
+}
 
+void
+os_context_set_fp_control(os_context_t *context, unsigned int value)
+{
+    context->uc_mcontext.fpregs->mxcsr = value ^ (0x3F << 7);
 }
 
 os_context_register_t *
@@ -231,21 +265,25 @@ os_flush_icache(os_vm_address_t __attribute__((unused)) address,
 #include <math.h>
 const long libm_anchor = (long)acos;
 
-#ifdef LISP_FEATURE_SW_INT_AVOIDANCE
+#if defined LISP_FEATURE_SW_INT_AVOIDANCE || defined LISP_FEATURE_PARTIAL_SW_INT_AVOIDANCE
 extern void sigtrap_handler();
 extern char* vm_thread_name(struct thread*);
 extern void sigset_tostring(const sigset_t*, char*, int);
-void synchronous_trap(lispobj* sp_at_interrupt, char* savearea)
+void synchronous_trap(char* savearea, lispobj* sp_at_interrupt)
 {
     os_context_t context;
     memset(&context, 0, sizeof context);
 
     // Create the signal context from the values pushed on the stack
     // by the lisp assembly routine.
+
+#ifndef LISP_FEATURE_PARTIAL_SW_INT_AVOIDANCE
+    // I don't know how to create the ucontext's representation from an XSAVE area
     context.uc_mcontext.fpregs = &context.__fpregs_mem;
     if (sizeof context.uc_mcontext.fpregs->_xmm[0].element != 16) lose("sigcontext size bug");
     memcpy(context.uc_mcontext.fpregs->_xmm[0].element, savearea, 16*16);
-    char* gprsave = savearea + 16*16;
+#endif
+    char* gprsave = savearea + 16*32; // 16 32-byte registers
     memcpy(context.uc_mcontext.gregs, gprsave, 15*8);
 
     context.uc_mcontext.gregs[REG_RSP] = (greg_t)sp_at_interrupt;
@@ -274,6 +312,10 @@ void synchronous_trap(lispobj* sp_at_interrupt, char* savearea)
             os_context_pc(&context), sp_at_interrupt, savearea,
             newmask_string); */
 
+#ifdef LISP_FEATURE_PARTIAL_SW_INT_AVOIDANCE
+    // This is for GC auto-triggers during pseudo-atomic, and nothing but
+    interrupt_handle_pending(&context);
+#else
     sigtrap_handler(0, 0, &context);
 
     if (context.uc_mcontext.gregs[REG_RSP] != (greg_t)sp_at_interrupt ||
@@ -284,6 +326,7 @@ void synchronous_trap(lispobj* sp_at_interrupt, char* savearea)
     // the return PC location that the assembly routine received.
     uword_t return_pc = context.uc_mcontext.gregs[REG_RIP];
     sp_at_interrupt[-1] = return_pc;
+#endif
 
     // act like a return-from-signal by restoring the signal mask
     // Ideally this would be performed in the asm routine only after restoring
