@@ -11,9 +11,11 @@
 
 (enable-test-parallelism)
 
-(assert (not (sb-kernel:member-type-p (sb-kernel:make-eql-type #\z))))
-(assert (not (sb-kernel:member-type-p (sb-kernel:make-eql-type 1.0))))
-(assert (sb-kernel:member-type-p (sb-kernel:make-eql-type -0.0s0)))
+(with-test (:name :make-eql-type)
+  (assert (not (sb-kernel:member-type-p (sb-kernel:make-eql-type #\z))))
+  (assert (not (sb-kernel:member-type-p (sb-kernel:make-eql-type 1.0))))
+  (assert (eql (sb-kernel:numeric-type-low (sb-kernel:make-eql-type -0f0)) -0f0))
+  (assert (eql (sb-kernel:numeric-type-high (sb-kernel:make-eql-type -0d0)) -0d0)))
 
 (with-test (:name (typexpand-1 typexpand typexpand-all :check-lexenv))
   (flet ((try (f) (assert-error (funcall f 'hash-table 3))))
@@ -528,15 +530,6 @@
                        sb-vm:*specialized-array-element-type-properties*))))
     (assert-tri-eq t t (subtypep 'array u))))
 
-(with-test (:name :bug-1258716)
-  (let ((intersection (sb-kernel:type-intersection
-                       (sb-kernel:specifier-type 'simple-vector)
-                       (sb-kernel:specifier-type `(vector #:unknown)))))
-    (assert (sb-kernel:array-type-p intersection))
-    ;; and not *wild-type*
-    (assert (sb-kernel:type= (sb-kernel:array-type-specialized-element-type intersection)
-                             sb-kernel:*universal-type*))))
-
 (with-test (:name :parse-safely)
   (dolist (x '(array integer cons))
     (assert (handler-case (sb-kernel:specifier-type `(,x . 0))
@@ -1043,3 +1036,97 @@
                 (sb-kernel::function-simple-union2-type-method type type)))
     (assert (eq type
                 (sb-kernel::function-simple-intersection2-type-method type type)))))
+
+(with-test (:name :array-canonical-difference)
+  (assert (eql
+           (specifier-type '(and (simple-array * (10)) (not string)))
+           (specifier-type '(and (simple-array * (10)) (not (simple-string 10))))))
+  (assert (type= (specifier-type '(or (simple-array fixnum) (not simple-array)))
+                 (specifier-type '(or (array fixnum) (not simple-array)))))
+  (assert (type= (specifier-type '(or (simple-array (mod 3)) (not simple-array)))
+                 (specifier-type '(or (array (mod 3)) (not simple-array)))))
+  (assert (subtypep '(or (array (mod 3)) (not simple-array))
+                    '(or (simple-array (mod 3)) (not simple-array))))
+  (assert (subtypep '(not (or (not simple-array) simple-string)) '(not string)))
+  (assert (subtypep '(and (array symbol) (not (and (array t) (not simple-array))))
+                    '(simple-array symbol)))
+  (assert (subtypep '(array t)
+                    '(or (and (array t) (not simple-array)) (simple-array integer))))
+  (assert (subtypep '(array (integer 15 27))
+                    '(or (and (array (integer 15 27)) (not simple-array))
+                      (simple-array (integer 17 30)))))
+  (assert (eql (specifier-type '(and (not (simple-array t)) (not (and (array t) (not simple-array)))))
+               (specifier-type '(not (array t)))))
+  (assert (subtypep '(and (not simple-array) (not (and (array t) (not simple-array))))
+                    '(and (not simple-array) (not (array t)))))
+  (assert (eql (specifier-type '(not simple-array))
+               (specifier-type '(or (not array) (and array (not simple-array))))))
+  (assert (eql (specifier-type '(or (and array (not (array t))) (and (array t) (not simple-array))))
+               (specifier-type '(and array (not (simple-array t))))))
+  (assert (eql (specifier-type '(or (not (vector * 10)) (and vector (not simple-array))))
+               (specifier-type '(not (simple-array * (10))))))
+  (assert (not (subtypep '(not (simple-array t))
+                         '(or (not array) (and (array t) (not simple-array))))))
+  (assert (eql (specifier-type '(or (not vector) (not (array t))))
+               (specifier-type '(not (vector t)))))
+  (assert (eql (specifier-type '(or (not simple-array) (not (array t))))
+               (specifier-type '(not (simple-array t)))))
+  (assert (eql (specifier-type '(or (and (not integer) (not vector) (not (array t))) vector (array t)))
+               (specifier-type '(not integer))))
+  (assert (eql (specifier-type '(or (and (not integer) (not vector) (not (array t))) vector))
+               (specifier-type '(or (and (not integer) (not (array t))) vector))))
+  (assert (not (subtypep '(not simple-bit-vector)
+                         '(or (not (array bit)) (and bit-vector (not simple-array))))))
+  (assert (eql (specifier-type '(not (and (array t) (not simple-array))))
+               (specifier-type '(or (not (array t)) simple-array))))
+  (assert (eql (specifier-type '(not (and base-string (not simple-array))))
+               (specifier-type '(or (not base-string) simple-base-string))))
+  (assert (eql (specifier-type '(and (array cons) (array number)))
+               (specifier-type '(array t))))
+  (assert (sb-kernel:intersection-type-p (specifier-type '(and (vector unknown) bit-vector)))))
+
+(with-test (:name :intersection-not-numeric)
+  (assert (eql
+           (specifier-type '(and (not (eql 1)) (not (eql 0))))
+           (specifier-type '(not bit))))
+  (assert (eql (specifier-type '(or (not integer) (not (rational 1 10))))
+               (specifier-type '(not (integer 1 10))))))
+
+(with-test (:name :class-canonical-difference)
+  (assert (eql
+           (specifier-type '(or (and stream standard-object) (and (not stream) standard-object)))
+           (specifier-type 'standard-object)))
+  (assert (eql
+           (specifier-type '(or (and stream standard-object) (not standard-object)))
+           (specifier-type '(or stream (not standard-object)))))
+  (assert (eql (specifier-type '(or (and standard-object sb-kernel:extended-sequence)
+                                 (and standard-object (not sb-kernel:extended-sequence))))
+               (specifier-type 'standard-object)))
+  (assert (eql (specifier-type '(or (and atom (not stream)) (and stream standard-object)))
+               (specifier-type '(or (and atom (not stream)) standard-object)))))
+
+(with-test (:name :cons-intersection)
+  (assert (eql (specifier-type '(and (cons (not array) atom) (cons (not integer) (not integer))))
+               (specifier-type '(cons (and (not array) (not integer)) (and atom (not integer))))))
+  (assert (eql (specifier-type '(or (and (not integer) (not stream) (not standard-object)) (and (not stream) standard-object)))
+               (specifier-type '(and (not integer) (not stream))))))
+
+(with-test (:name :float-zero-typep)
+  (checked-compile-and-assert
+      ()
+      `(lambda (x)
+         (typep x '(single-float * (0.0))))
+    ((0.0) nil)
+    ((-0.0) nil)
+    ((1.0) nil)
+    ((-1.0) t))
+  (checked-compile-and-assert
+      ()
+      `(lambda (x)
+         (typep x '(or (member 0.0) (single-float (0.0) 1.0))))
+    ((0.0) t)
+    ((1.0) t)
+    ((2.0) nil)
+    ((-0.0) nil)
+    ((-1.0) nil)
+    ((-2.0) nil)))
