@@ -942,7 +942,7 @@
              (assert (not (loop for line in (disassembly-lines negative-test)
                                 thereis (search telltale line)))))))
 
-(with-test (:name :bash-copiers-byte-or-larger)
+(with-test (:name :bash-copiers-byte-or-larger :skipped-on :sb-devel)
   (dolist (f '(sb-kernel::ub8-bash-copy
                sb-kernel::ub16-bash-copy
                sb-kernel::ub32-bash-copy
@@ -1417,3 +1417,11 @@
   ;; four raw words: jump table count word, two user data words, and a padding word
   (assert (= 32 (sb-kernel:code-n-unboxed-data-bytes
                  (sb-kernel:fun-code-header #'different-constants)))))
+
+(with-test (:name :push-cons)
+  (let ((f (checked-compile
+            '(lambda (alist x y)
+              (declare (optimize (sb-c:instrument-consing 0)))
+              (push (cons x y) alist)))))
+    ;; should have 1 call to list-alloc-tramp, not one for the cons of x, y and one for push
+    (assert (= 1 (count 'sb-c:call (get-simple-fun-instruction-model f) :key 'second)))))
